@@ -63,7 +63,12 @@ class AllocationBuilder final {
 
   // Compound store of a constant into a field.
   void Store(const FieldAccess& access, ObjectRef value) {
-    Store(access, jsgraph()->ConstantNoHole(value, broker_));
+    if (access.machine_type == MachineType::IndirectPointer()) {
+      Store(access,
+            jsgraph()->TrustedHeapConstant(value.AsHeapObject().object()));
+    } else {
+      Store(access, jsgraph()->ConstantNoHole(value, broker_));
+    }
   }
 
   void FinishAndChange(Node* node) {
@@ -81,7 +86,7 @@ class AllocationBuilder final {
  protected:
   JSGraph* jsgraph() { return jsgraph_; }
   Isolate* isolate() const { return jsgraph_->isolate(); }
-  Graph* graph() { return jsgraph_->graph(); }
+  TFGraph* graph() { return jsgraph_->graph(); }
   CommonOperatorBuilder* common() { return jsgraph_->common(); }
   SimplifiedOperatorBuilder* simplified() { return jsgraph_->simplified(); }
 
