@@ -16,15 +16,16 @@
 #include "src/compiler/backend/instruction-codes.h"
 #include "src/compiler/common-operator.h"
 #include "src/compiler/frame-states.h"
-#include "src/compiler/graph.h"
 #include "src/compiler/node.h"
 #include "src/compiler/schedule.h"
+#include "src/compiler/turbofan-graph.h"
 #include "src/compiler/turboshaft/graph.h"
 #include "src/compiler/turboshaft/loop-finder.h"
 #include "src/compiler/turboshaft/operations.h"
 #include "src/deoptimizer/deoptimizer.h"
 #include "src/execution/frames.h"
 #include "src/execution/isolate-utils-inl.h"
+#include "src/objects/heap-object-inl.h"
 #include "src/objects/instance-type-inl.h"
 #include "src/utils/ostreams.h"
 
@@ -261,6 +262,9 @@ std::ostream& operator<<(std::ostream& os, const InstructionOperand& op) {
         case MachineRepresentation::kWord64:
           os << "|w64";
           break;
+        case MachineRepresentation::kFloat16:
+          os << "|f16";
+          break;
         case MachineRepresentation::kFloat32:
           os << "|f32";
           break;
@@ -288,6 +292,9 @@ std::ostream& operator<<(std::ostream& os, const InstructionOperand& op) {
         case MachineRepresentation::kCompressed:
           os << "|c";
           break;
+        case MachineRepresentation::kProtectedPointer:
+          os << "|pp";
+          break;
         case MachineRepresentation::kIndirectPointer:
           os << "|ip";
           break;
@@ -295,6 +302,7 @@ std::ostream& operator<<(std::ostream& os, const InstructionOperand& op) {
           os << "|sb";
           break;
         case MachineRepresentation::kMapWord:
+        case MachineRepresentation::kFloat16RawBits:
           UNREACHABLE();
       }
       return os << "]";
@@ -481,6 +489,10 @@ std::ostream& operator<<(std::ostream& os, const FlagsMode& fm) {
       return os << "trap";
     case kFlags_select:
       return os << "select";
+    case kFlags_conditional_set:
+      return os << "conditional set";
+    case kFlags_conditional_branch:
+      return os << "conditional branch";
   }
   UNREACHABLE();
 }
@@ -594,16 +606,17 @@ Constant::Constant(RelocatablePtrConstantInfo info) {
   rmode_ = info.rmode();
 }
 
-Handle<HeapObject> Constant::ToHeapObject() const {
+IndirectHandle<HeapObject> Constant::ToHeapObject() const {
   DCHECK(kHeapObject == type() || kCompressedHeapObject == type());
-  Handle<HeapObject> value(
+  IndirectHandle<HeapObject> value(
       reinterpret_cast<Address*>(static_cast<intptr_t>(value_)));
   return value;
 }
 
-Handle<Code> Constant::ToCode() const {
+IndirectHandle<Code> Constant::ToCode() const {
   DCHECK_EQ(kHeapObject, type());
-  Handle<Code> value(reinterpret_cast<Address*>(static_cast<intptr_t>(value_)));
+  IndirectHandle<Code> value(
+      reinterpret_cast<Address*>(static_cast<intptr_t>(value_)));
   DCHECK(IsCode(*value));
   return value;
 }
@@ -619,7 +632,7 @@ std::ostream& operator<<(std::ostream& os, const Constant& constant) {
     case Constant::kFloat64:
       return os << constant.ToFloat64().value();
     case Constant::kExternalReference:
-      return os << constant.ToExternalReference().address();
+      return os << constant.ToExternalReference();
     case Constant::kHeapObject:  // Fall through.
     case Constant::kCompressedHeapObject:
       return os << Brief(*constant.ToHeapObject());
@@ -663,7 +676,7 @@ InstructionBlock::InstructionBlock(Zone* zone, RpoNumber rpo_number,
       switch_target_(false),
       code_target_alignment_(false),
       loop_header_alignment_(false),
-      needs_frame_(false),
+      needs_frame_(!v8_flags.turbo_elide_frames),
       must_construct_frame_(false),
       must_deconstruct_frame_(false),
       omitted_by_jump_threading_(false) {}
@@ -708,7 +721,7 @@ static InstructionBlock* InstructionBlockFor(Zone* zone,
   InstructionBlock* instr_block = zone->New<InstructionBlock>(
       zone, GetRpo(block), GetRpo(block->loop_header()), GetLoopEndRpo(block),
       GetRpo(block->dominator()), block->deferred(), is_handler);
-  // Map successors and precessors
+  // Map successors and predecessors
   instr_block->successors().reserve(block->SuccessorCount());
   for (BasicBlock* successor : block->successors()) {
     instr_block->successors().push_back(GetRpo(successor));
@@ -1030,6 +1043,8 @@ static MachineRepresentation FilterRepresentation(MachineRepresentation rep) {
     case MachineRepresentation::kWord8:
     case MachineRepresentation::kWord16:
       return InstructionSequence::DefaultRepresentation();
+    case MachineRepresentation::kFloat16:
+      return MachineRepresentation::kFloat32;
     case MachineRepresentation::kWord32:
     case MachineRepresentation::kWord64:
     case MachineRepresentation::kTaggedSigned:
@@ -1041,15 +1056,15 @@ static MachineRepresentation FilterRepresentation(MachineRepresentation rep) {
     case MachineRepresentation::kSimd256:
     case MachineRepresentation::kCompressedPointer:
     case MachineRepresentation::kCompressed:
+    case MachineRepresentation::kProtectedPointer:
     case MachineRepresentation::kSandboxedPointer:
       return rep;
     case MachineRepresentation::kNone:
     case MachineRepresentation::kMapWord:
     case MachineRepresentation::kIndirectPointer:
-      break;
+    case MachineRepresentation::kFloat16RawBits:
+      UNREACHABLE();
   }
-
-  UNREACHABLE();
 }
 
 MachineRepresentation InstructionSequence::GetRepresentation(
@@ -1143,7 +1158,8 @@ namespace {
 size_t GetConservativeFrameSizeInBytes(FrameStateType type,
                                        size_t parameters_count,
                                        size_t locals_count,
-                                       BytecodeOffset bailout_id) {
+                                       BytecodeOffset bailout_id,
+                                       uint32_t wasm_liftoff_frame_size) {
   switch (type) {
     case FrameStateType::kUnoptimizedFunction: {
       auto info = UnoptimizedFrameInfo::Conservative(
@@ -1183,6 +1199,10 @@ size_t GetConservativeFrameSizeInBytes(FrameStateType type,
           config);
       return info.frame_size_in_bytes();
     }
+#if V8_ENABLE_WEBASSEMBLY
+    case FrameStateType::kLiftoffFunction:
+      return wasm_liftoff_frame_size;
+#endif  // V8_ENABLE_WEBASSEMBLY
   }
   UNREACHABLE();
 }
@@ -1191,13 +1211,14 @@ size_t GetTotalConservativeFrameSizeInBytes(FrameStateType type,
                                             size_t parameters_count,
                                             size_t locals_count,
                                             BytecodeOffset bailout_id,
+                                            uint32_t wasm_liftoff_frame_size,
                                             FrameStateDescriptor* outer_state) {
   size_t outer_total_conservative_frame_size_in_bytes =
       (outer_state == nullptr)
           ? 0
           : outer_state->total_conservative_frame_size_in_bytes();
   return GetConservativeFrameSizeInBytes(type, parameters_count, locals_count,
-                                         bailout_id) +
+                                         bailout_id, wasm_liftoff_frame_size) +
          outer_total_conservative_frame_size_in_bytes;
 }
 
@@ -1205,22 +1226,28 @@ size_t GetTotalConservativeFrameSizeInBytes(FrameStateType type,
 
 FrameStateDescriptor::FrameStateDescriptor(
     Zone* zone, FrameStateType type, BytecodeOffset bailout_id,
-    OutputFrameStateCombine state_combine, size_t parameters_count,
-    size_t locals_count, size_t stack_count,
-    MaybeHandle<SharedFunctionInfo> shared_info,
-    FrameStateDescriptor* outer_state)
+    OutputFrameStateCombine state_combine, uint16_t parameters_count,
+    uint16_t max_arguments, size_t locals_count, size_t stack_count,
+    MaybeIndirectHandle<SharedFunctionInfo> shared_info,
+    MaybeIndirectHandle<BytecodeArray> bytecode_array,
+    FrameStateDescriptor* outer_state, uint32_t wasm_liftoff_frame_size,
+    uint32_t wasm_function_index)
     : type_(type),
       bailout_id_(bailout_id),
       frame_state_combine_(state_combine),
       parameters_count_(parameters_count),
+      max_arguments_(max_arguments),
       locals_count_(locals_count),
       stack_count_(stack_count),
       total_conservative_frame_size_in_bytes_(
           GetTotalConservativeFrameSizeInBytes(
-              type, parameters_count, locals_count, bailout_id, outer_state)),
+              type, parameters_count, locals_count, bailout_id,
+              wasm_liftoff_frame_size, outer_state)),
       values_(zone),
       shared_info_(shared_info),
-      outer_state_(outer_state) {}
+      bytecode_array_(bytecode_array),
+      outer_state_(outer_state),
+      wasm_function_index_(wasm_function_index) {}
 
 size_t FrameStateDescriptor::GetHeight() const {
   switch (type()) {
@@ -1245,6 +1272,10 @@ size_t FrameStateDescriptor::GetHeight() const {
       //   CreateJavaScriptBuiltinContinuationFrameState), and
       // - does *not* include the context.
       return parameters_count();
+#if V8_ENABLE_WEBASSEMBLY
+    case FrameStateType::kLiftoffFunction:
+      return locals_count() + parameters_count();
+#endif
   }
   UNREACHABLE();
 }
@@ -1286,13 +1317,13 @@ size_t FrameStateDescriptor::GetJSFrameCount() const {
 #if V8_ENABLE_WEBASSEMBLY
 JSToWasmFrameStateDescriptor::JSToWasmFrameStateDescriptor(
     Zone* zone, FrameStateType type, BytecodeOffset bailout_id,
-    OutputFrameStateCombine state_combine, size_t parameters_count,
+    OutputFrameStateCombine state_combine, uint16_t parameters_count,
     size_t locals_count, size_t stack_count,
-    MaybeHandle<SharedFunctionInfo> shared_info,
-    FrameStateDescriptor* outer_state, const wasm::FunctionSig* wasm_signature)
+    MaybeIndirectHandle<SharedFunctionInfo> shared_info,
+    FrameStateDescriptor* outer_state, const wasm::CanonicalSig* wasm_signature)
     : FrameStateDescriptor(zone, type, bailout_id, state_combine,
-                           parameters_count, locals_count, stack_count,
-                           shared_info, outer_state),
+                           parameters_count, 0, locals_count, stack_count,
+                           shared_info, {}, outer_state),
       return_kind_(wasm::WasmReturnTypeFromSignature(wasm_signature)) {}
 #endif  // V8_ENABLE_WEBASSEMBLY
 
@@ -1323,20 +1354,25 @@ std::ostream& operator<<(std::ostream& os, StateValueKind kind) {
       return os << "ArgumentsElements";
     case StateValueKind::kArgumentsLength:
       return os << "ArgumentsLength";
+    case StateValueKind::kRestLength:
+      return os << "RestLength";
     case StateValueKind::kPlain:
       return os << "Plain";
     case StateValueKind::kOptimizedOut:
       return os << "OptimizedOut";
-    case StateValueKind::kNested:
-      return os << "Nested";
+    case StateValueKind::kNestedObject:
+      return os << "NestedObject";
     case StateValueKind::kDuplicate:
       return os << "Duplicate";
+    case StateValueKind::kStringConcat:
+      return os << "StringConcat";
   }
 }
 
 void StateValueDescriptor::Print(std::ostream& os) const {
   os << "kind=" << kind_ << ", type=" << type_;
-  if (kind_ == StateValueKind::kDuplicate || kind_ == StateValueKind::kNested) {
+  if (kind_ == StateValueKind::kDuplicate ||
+      kind_ == StateValueKind::kNestedObject) {
     os << ", id=" << id_;
   } else if (kind_ == StateValueKind::kArgumentsElements) {
     os << ", args_type=" << args_type_;
