@@ -5,19 +5,19 @@
 #ifndef V8_HEAP_LOCAL_FACTORY_INL_H_
 #define V8_HEAP_LOCAL_FACTORY_INL_H_
 
-#include "src/heap/factory-base-inl.h"
 #include "src/heap/local-factory.h"
+// Include the non-inl header before the rest of the headers.
+
+#include "src/heap/factory-base-inl.h"
 #include "src/roots/roots-inl.h"
 
 namespace v8 {
 namespace internal {
 
-#define ACCESSOR_INFO_ACCESSOR(Type, name, CamelName)                          \
-  Handle<Type> LocalFactory::name() {                                          \
-    /* Do a bit of handle location magic to cast the Handle without having */  \
-    /* to pull in Type::cast. We know the type is right by construction.   */  \
-    return Handle<Type>(                                                       \
-        isolate()->isolate_->root_handle(RootIndex::k##CamelName).location()); \
+#define ACCESSOR_INFO_ACCESSOR(Type, name, CamelName)               \
+  DirectHandle<Type> LocalFactory::name() {                         \
+    return UncheckedCast<Type>(                                     \
+        isolate()->isolate_->root_handle(RootIndex::k##CamelName)); \
   }
 ACCESSOR_INFO_ROOT_LIST(ACCESSOR_INFO_ACCESSOR)
 #undef ACCESSOR_INFO_ACCESSOR
