@@ -15,9 +15,9 @@ namespace v8_inspector {
 const int kMaxCustomPreviewDepth = 20;
 
 void generateCustomPreview(
-    int sessionId, const String16& groupName, v8::Local<v8::Object> object,
-    v8::MaybeLocal<v8::Value> config, int maxDepth,
-    std::unique_ptr<protocol::Runtime::CustomPreview>* preview);
+    v8::Isolate* isolate, int sessionId, const String16& groupName,
+    v8::Local<v8::Object> object, v8::MaybeLocal<v8::Value> config,
+    int maxDepth, std::unique_ptr<protocol::Runtime::CustomPreview>* preview);
 
 }  // namespace v8_inspector
 
