@@ -2,15 +2,17 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
+#ifndef V8_OBJECTS_JS_LOCALE_INL_H_
+#define V8_OBJECTS_JS_LOCALE_INL_H_
+
 #ifndef V8_INTL_SUPPORT
 #error Internationalization is expected to be enabled.
 #endif  // V8_INTL_SUPPORT
 
-#ifndef V8_OBJECTS_JS_LOCALE_INL_H_
-#define V8_OBJECTS_JS_LOCALE_INL_H_
+#include "src/objects/js-locale.h"
+// Include the non-inl header before the rest of the headers.
 
 #include "src/api/api-inl.h"
-#include "src/objects/js-locale.h"
 #include "src/objects/objects-inl.h"
 
 // Has to be the last include (doesn't have include guards):
@@ -23,7 +25,7 @@ namespace internal {
 
 TQ_OBJECT_CONSTRUCTORS_IMPL(JSLocale)
 
-ACCESSORS(JSLocale, icu_locale, Managed<icu::Locale>, kIcuLocaleOffset)
+ACCESSORS(JSLocale, icu_locale, Tagged<Managed<icu::Locale>>, kIcuLocaleOffset)
 
 }  // namespace internal
 }  // namespace v8
