@@ -8,8 +8,7 @@
 #include "src/objects/map.h"
 #include "src/objects/objects.h"
 
-namespace v8 {
-namespace internal {
+namespace v8::internal {
 
 // This is the base class for object's body descriptors.
 //
@@ -71,6 +70,11 @@ class BodyDescriptorBase {
   template <typename ObjectVisitor>
   static inline void IterateProtectedPointer(Tagged<HeapObject> obj, int offset,
                                              ObjectVisitor* v);
+#ifdef V8_ENABLE_LEAPTIERING
+  template <typename ObjectVisitor>
+  static inline void IterateJSDispatchEntry(Tagged<HeapObject> obj, int offset,
+                                            ObjectVisitor* v);
+#endif  // V8_ENABLE_LEAPTIERING
 
  protected:
   // Returns true for all header and embedder fields.
@@ -78,12 +82,32 @@ class BodyDescriptorBase {
                                                      Tagged<HeapObject> obj,
                                                      int offset);
 
-  // Treats all header and embedder fields in the range as tagged.
+  // Treats all header and in-object fields in the range as tagged. Figures out
+  // dynamically whether the object has embedder fields and visits them
+  // accordingly (as tagged fields and as external pointers).
   template <typename ObjectVisitor>
   static inline void IterateJSObjectBodyImpl(Tagged<Map> map,
                                              Tagged<HeapObject> obj,
                                              int start_offset, int end_offset,
                                              ObjectVisitor* v);
+
+  // Treats all header and in-object fields in the range as tagged.
+  template <typename ObjectVisitor>
+  static inline void IterateJSObjectBodyWithoutEmbedderFieldsImpl(
+      Tagged<Map> map, Tagged<HeapObject> obj, int start_offset, int end_offset,
+      ObjectVisitor* v);
+};
+
+// This class describes a body of an object without any pointers.
+class DataOnlyBodyDescriptor : public BodyDescriptorBase {
+ public:
+  template <typename ObjectVisitor>
+  static inline void IterateBody(Tagged<Map> map, Tagged<HeapObject> obj,
+                                 int object_size, ObjectVisitor* v) {}
+
+ private:
+  // Note: {SizeOf} is not implemented here; sub-classes will have to implement
+  // it.
 };
 
 // This class describes a body of an object in which all pointer fields are
@@ -107,11 +131,8 @@ class FixedRangeBodyDescriptor : public BodyDescriptorBase {
     IterateBody(map, obj, v);
   }
 
- private:
-  static inline int SizeOf(Tagged<Map> map, Tagged<HeapObject> object) {
-    // Has to be implemented by the subclass.
-    UNREACHABLE();
-  }
+  // Note: {SizeOf} is not implemented here; sub-classes will have to implement
+  // it.
 };
 
 // This class describes a body of an object of a fixed size
@@ -120,14 +141,23 @@ class FixedRangeBodyDescriptor : public BodyDescriptorBase {
 // All pointers have to be strong.
 template <int start_offset, int end_offset, int size>
 class FixedBodyDescriptor
-    : public FixedRangeBodyDescriptor<start_offset, end_offset> {
+    : public std::conditional_t<
+          start_offset == end_offset, DataOnlyBodyDescriptor,
+          FixedRangeBodyDescriptor<start_offset, end_offset>> {
  public:
-  static const int kSize = size;
+  static constexpr int kSize = size;
+
   static inline int SizeOf(Tagged<Map> map, Tagged<HeapObject> object) {
+    DCHECK_EQ(kSize, map->instance_size());
     return kSize;
   }
 };
 
+template <typename T>
+using FixedBodyDescriptorFor =
+    FixedBodyDescriptor<T::kStartOfStrongFieldsOffset,
+                        T::kEndOfStrongFieldsOffset, T::kSize>;
+
 // This class describes a body of an object in which all pointer fields are
 // located in the [start_offset, object_size) interval.
 // All pointers have to be strong.
@@ -142,11 +172,8 @@ class SuffixRangeBodyDescriptor : public BodyDescriptorBase {
     IteratePointers(obj, start_offset, object_size, v);
   }
 
- private:
-  static inline int SizeOf(Tagged<Map> map, Tagged<HeapObject> object) {
-    // Has to be implemented by the subclass.
-    UNREACHABLE();
-  }
+  // Note: {SizeOf} is not implemented here; sub-classes will have to implement
+  // it.
 };
 
 // This class describes a body of an object of a variable size
@@ -159,11 +186,9 @@ class FlexibleBodyDescriptor : public SuffixRangeBodyDescriptor<start_offset> {
   static inline int SizeOf(Tagged<Map> map, Tagged<HeapObject> object);
 };
 
-// A forward-declacable descriptor body alias for most of the Struct successors.
+// A forward-declarable descriptor body alias for most of the Struct successors.
 class StructBodyDescriptor
-    : public FlexibleBodyDescriptor<HeapObject::kHeaderSize> {
- public:
-};
+    : public FlexibleBodyDescriptor<HeapObject::kHeaderSize> {};
 
 // This class describes a body of an object in which all pointer fields are
 // located in the [start_offset, object_size) interval.
@@ -179,11 +204,8 @@ class SuffixRangeWeakBodyDescriptor : public BodyDescriptorBase {
     IterateMaybeWeakPointers(obj, start_offset, object_size, v);
   }
 
- private:
-  static inline int SizeOf(Tagged<Map> map, Tagged<HeapObject> object) {
-    // Has to be implemented by the subclass.
-    UNREACHABLE();
-  }
+  // Note: {SizeOf} is not implemented here; sub-classes will have to implement
+  // it.
 };
 
 // This class describes a body of an object of a variable size
@@ -197,27 +219,13 @@ class FlexibleWeakBodyDescriptor
   static inline int SizeOf(Tagged<Map> map, Tagged<HeapObject> object);
 };
 
-// This class describes a body of an object without any pointers.
-class DataOnlyBodyDescriptor : public BodyDescriptorBase {
- public:
-  template <typename ObjectVisitor>
-  static inline void IterateBody(Tagged<Map> map, Tagged<HeapObject> obj,
-                                 int object_size, ObjectVisitor* v) {}
-
- private:
-  static inline int SizeOf(Tagged<Map> map, Tagged<HeapObject> object) {
-    // Has to be implemented by the subclass.
-    UNREACHABLE();
-  }
-};
-
 // This class describes a body of an object which has a parent class that also
 // has a body descriptor. This represents a union of the parent's body
 // descriptor, and a new descriptor for the child -- so, both parent and child's
 // slots are iterated. The parent must be fixed size, and its slots be disjoint
 // with the child's.
 template <class ParentBodyDescriptor, class ChildBodyDescriptor>
-class SubclassBodyDescriptor final : public BodyDescriptorBase {
+class SubclassBodyDescriptor : public BodyDescriptorBase {
  public:
   // The parent must end be before the child's start offset, to make sure that
   // their slots are disjoint.
@@ -244,7 +252,86 @@ class SubclassBodyDescriptor final : public BodyDescriptorBase {
   }
 };
 
-}  // namespace internal
-}  // namespace v8
+// Visitor for exposed trusted objects with fixed layout according to
+// FixedBodyDescriptor.
+template <typename T, IndirectPointerTag kTag>
+class FixedExposedTrustedObjectBodyDescriptor
+    : public FixedBodyDescriptorFor<T> {
+  static_assert(std::is_base_of_v<ExposedTrustedObject, T>);
+  using Base = FixedBodyDescriptorFor<T>;
+
+ public:
+  template <typename ObjectVisitor>
+  static inline void IterateBody(Tagged<Map> map, Tagged<HeapObject> obj,
+                                 int object_size, ObjectVisitor* v) {
+    Base::IterateSelfIndirectPointer(obj, kTag, v);
+    Base::IterateBody(map, obj, object_size, v);
+  }
+};
+
+// A mix-in for visiting a trusted pointer field.
+template <size_t kFieldOffset, IndirectPointerTag kTag>
+struct WithStrongTrustedPointer {
+  template <typename Base>
+  class BodyDescriptor : public Base {
+   public:
+    template <typename ObjectVisitor>
+    static inline void IterateBody(Tagged<Map> map, Tagged<HeapObject> obj,
+                                   int object_size, ObjectVisitor* v) {
+      Base::IterateBody(map, obj, object_size, v);
+      Base::IterateTrustedPointer(obj, kFieldOffset, v,
+                                  IndirectPointerMode::kStrong, kTag);
+    }
+  };
+};
+
+template <size_t kFieldOffset>
+using WithStrongCodePointer =
+    WithStrongTrustedPointer<kFieldOffset, kCodeIndirectPointerTag>;
+
+// A mix-in for visiting an external pointer field.
+template <size_t kFieldOffset, ExternalPointerTagRange kTagRange>
+struct WithExternalPointer {
+  template <typename Base>
+  class BodyDescriptor : public Base {
+   public:
+    template <typename ObjectVisitor>
+    static inline void IterateBody(Tagged<Map> map, Tagged<HeapObject> obj,
+                                   int object_size, ObjectVisitor* v) {
+      Base::IterateBody(map, obj, object_size, v);
+      v->VisitExternalPointer(
+          obj, obj->RawExternalPointerField(kFieldOffset, kTagRange));
+    }
+  };
+};
+
+// A mix-in for visiting an external pointer field.
+template <size_t kFieldOffset>
+struct WithProtectedPointer {
+  template <typename Base>
+  class BodyDescriptor : public Base {
+   public:
+    template <typename ObjectVisitor>
+    static inline void IterateBody(Tagged<Map> map, Tagged<HeapObject> obj,
+                                   int object_size, ObjectVisitor* v) {
+      Base::IterateBody(map, obj, object_size, v);
+      Base::IterateProtectedPointer(obj, kFieldOffset, v);
+    }
+  };
+};
+
+// Stack multiple body descriptors; the first template argument is the base,
+// followed by mix-ins.
+template <typename Base, typename FirstMixin, typename... MoreMixins>
+class StackedBodyDescriptor
+    : public StackedBodyDescriptor<
+          typename FirstMixin::template BodyDescriptor<Base>, MoreMixins...> {};
+
+// Define a specialization for the base case of only one mixin.
+template <typename Base, typename FirstMixin>
+class StackedBodyDescriptor<Base, FirstMixin>
+    : public FirstMixin::template BodyDescriptor<Base> {};
+
+}  // namespace v8::internal
 
 #endif  // V8_OBJECTS_OBJECTS_BODY_DESCRIPTORS_H_
