@@ -9,87 +9,51 @@
 #include "src/objects/objects-inl.h"
 #include "src/objects/slots.h"
 #include "src/objects/smi.h"
+#include "src/runtime/runtime-utils.h"
 #include "src/strings/string-builder-inl.h"
 #include "src/strings/unicode-inl.h"
 
-#if V8_ENABLE_WEBASSEMBLY
-// TODO(chromium:1236668): Drop this when the "SaveAndClearThreadInWasmFlag"
-// approach is no longer needed.
-#include "src/trap-handler/trap-handler.h"
-#endif  // V8_ENABLE_WEBASSEMBLY
-
 namespace v8 {
 namespace internal {
 
-namespace {
-
-#if V8_ENABLE_WEBASSEMBLY
-class V8_NODISCARD SaveAndClearThreadInWasmFlag {
- public:
-  explicit SaveAndClearThreadInWasmFlag(Isolate* isolate) : isolate_(isolate) {
-    if (trap_handler::IsTrapHandlerEnabled()) {
-      if (trap_handler::IsThreadInWasm()) {
-        thread_was_in_wasm_ = true;
-        trap_handler::ClearThreadInWasm();
-      }
-    }
-  }
-  ~SaveAndClearThreadInWasmFlag() {
-    if (thread_was_in_wasm_ && !isolate_->has_exception()) {
-      trap_handler::SetThreadInWasm();
-    }
-  }
-
- private:
-  bool thread_was_in_wasm_{false};
-  Isolate* isolate_;
-};
-#define CLEAR_THREAD_IN_WASM_SCOPE \
-  SaveAndClearThreadInWasmFlag non_wasm_scope(isolate)
-#else
-#define CLEAR_THREAD_IN_WASM_SCOPE (void)0
-#endif  // V8_ENABLE_WEBASSEMBLY
-
-}  // namespace
-
 RUNTIME_FUNCTION(Runtime_GetSubstitution) {
   HandleScope scope(isolate);
   DCHECK_EQ(5, args.length());
-  Handle<String> matched = args.at<String>(0);
-  Handle<String> subject = args.at<String>(1);
+  DirectHandle<String> matched = args.at<String>(0);
+  DirectHandle<String> subject = args.at<String>(1);
   int position = args.smi_value_at(2);
-  Handle<String> replacement = args.at<String>(3);
+  DirectHandle<String> replacement = args.at<String>(3);
   int start_index = args.smi_value_at(4);
 
   // A simple match without captures.
   class SimpleMatch : public String::Match {
    public:
-    SimpleMatch(Handle<String> match, Handle<String> prefix,
-                Handle<String> suffix)
+    SimpleMatch(DirectHandle<String> match, DirectHandle<String> prefix,
+                DirectHandle<String> suffix)
         : match_(match), prefix_(prefix), suffix_(suffix) {}
 
-    Handle<String> GetMatch() override { return match_; }
-    Handle<String> GetPrefix() override { return prefix_; }
-    Handle<String> GetSuffix() override { return suffix_; }
+    DirectHandle<String> GetMatch() override { return match_; }
+    DirectHandle<String> GetPrefix() override { return prefix_; }
+    DirectHandle<String> GetSuffix() override { return suffix_; }
 
     int CaptureCount() override { return 0; }
     bool HasNamedCaptures() override { return false; }
-    MaybeHandle<String> GetCapture(int i, bool* capture_exists) override {
+    MaybeDirectHandle<String> GetCapture(int i, bool* capture_exists) override {
       *capture_exists = false;
       return match_;  // Return arbitrary string handle.
     }
-    MaybeHandle<String> GetNamedCapture(Handle<String> name,
-                                        CaptureState* state) override {
+    MaybeDirectHandle<String> GetNamedCapture(DirectHandle<String> name,
+                                              CaptureState* state) override {
       UNREACHABLE();
     }
 
    private:
-    Handle<String> match_, prefix_, suffix_;
+    DirectHandle<String> match_, prefix_, suffix_;
   };
 
-  Handle<String> prefix =
+  DirectHandle<String> prefix =
       isolate->factory()->NewSubString(subject, 0, position);
-  Handle<String> suffix = isolate->factory()->NewSubString(
+  DirectHandle<String> suffix = isolate->factory()->NewSubString(
       subject, position + matched->length(), subject->length());
   SimpleMatch match(matched, prefix, suffix);
 
@@ -98,32 +62,32 @@ RUNTIME_FUNCTION(Runtime_GetSubstitution) {
       String::GetSubstitution(isolate, &match, replacement, start_index));
 }
 
-// This may return an empty MaybeHandle if an exception is thrown or
+// This may return an empty MaybeDirectHandle if an exception is thrown or
 // we abort due to reaching the recursion limit.
-MaybeHandle<String> StringReplaceOneCharWithString(
-    Isolate* isolate, Handle<String> subject, Handle<String> search,
-    Handle<String> replace, bool* found, int recursion_limit) {
+MaybeDirectHandle<String> StringReplaceOneCharWithString(
+    Isolate* isolate, DirectHandle<String> subject, DirectHandle<String> search,
+    DirectHandle<String> replace, bool* found, int recursion_limit) {
   StackLimitCheck stackLimitCheck(isolate);
   if (stackLimitCheck.HasOverflowed() || (recursion_limit == 0)) {
-    return MaybeHandle<String>();
+    return MaybeDirectHandle<String>();
   }
   recursion_limit--;
   if (IsConsString(*subject)) {
-    Tagged<ConsString> cons = ConsString::cast(*subject);
-    Handle<String> first = handle(cons->first(), isolate);
-    Handle<String> second = handle(cons->second(), isolate);
-    Handle<String> new_first;
+    Tagged<ConsString> cons = Cast<ConsString>(*subject);
+    DirectHandle<String> first(cons->first(), isolate);
+    DirectHandle<String> second(cons->second(), isolate);
+    DirectHandle<String> new_first;
     if (!StringReplaceOneCharWithString(isolate, first, search, replace, found,
                                         recursion_limit).ToHandle(&new_first)) {
-      return MaybeHandle<String>();
+      return MaybeDirectHandle<String>();
     }
     if (*found) return isolate->factory()->NewConsString(new_first, second);
 
-    Handle<String> new_second;
+    DirectHandle<String> new_second;
     if (!StringReplaceOneCharWithString(isolate, second, search, replace, found,
                                         recursion_limit)
              .ToHandle(&new_second)) {
-      return MaybeHandle<String>();
+      return MaybeDirectHandle<String>();
     }
     if (*found) return isolate->factory()->NewConsString(first, new_second);
 
@@ -132,12 +96,12 @@ MaybeHandle<String> StringReplaceOneCharWithString(
     int index = String::IndexOf(isolate, subject, search, 0);
     if (index == -1) return subject;
     *found = true;
-    Handle<String> first = isolate->factory()->NewSubString(subject, 0, index);
-    Handle<String> cons1;
+    DirectHandle<String> first =
+        isolate->factory()->NewSubString(subject, 0, index);
+    DirectHandle<String> cons1;
     ASSIGN_RETURN_ON_EXCEPTION(
-        isolate, cons1, isolate->factory()->NewConsString(first, replace),
-        String);
-    Handle<String> second =
+        isolate, cons1, isolate->factory()->NewConsString(first, replace));
+    DirectHandle<String> second =
         isolate->factory()->NewSubString(subject, index + 1, subject->length());
     return isolate->factory()->NewConsString(cons1, second);
   }
@@ -146,15 +110,15 @@ MaybeHandle<String> StringReplaceOneCharWithString(
 RUNTIME_FUNCTION(Runtime_StringReplaceOneCharWithString) {
   HandleScope scope(isolate);
   DCHECK_EQ(3, args.length());
-  Handle<String> subject = args.at<String>(0);
-  Handle<String> search = args.at<String>(1);
-  Handle<String> replace = args.at<String>(2);
+  DirectHandle<String> subject = args.at<String>(0);
+  DirectHandle<String> search = args.at<String>(1);
+  DirectHandle<String> replace = args.at<String>(2);
 
   // If the cons string tree is too deep, we simply abort the recursion and
   // retry with a flattened subject string.
   const int kRecursionLimit = 0x1000;
   bool found = false;
-  Handle<String> result;
+  DirectHandle<String> result;
   if (StringReplaceOneCharWithString(isolate, subject, search, replace, &found,
                                      kRecursionLimit).ToHandle(&result)) {
     return *result;
@@ -180,7 +144,7 @@ RUNTIME_FUNCTION(Runtime_StringLastIndexOf) {
 RUNTIME_FUNCTION(Runtime_StringSubstring) {
   HandleScope scope(isolate);
   DCHECK_EQ(3, args.length());
-  Handle<String> string = args.at<String>(0);
+  DirectHandle<String> string = args.at<String>(0);
   int start = args.smi_value_at(1);
   int end = args.smi_value_at(2);
   DCHECK_LE(0, start);
@@ -190,12 +154,12 @@ RUNTIME_FUNCTION(Runtime_StringSubstring) {
 }
 
 RUNTIME_FUNCTION(Runtime_StringAdd) {
-  // This is used by Wasm stringrefs.
-  CLEAR_THREAD_IN_WASM_SCOPE;
+  // This is used by Wasm.
+  SaveAndClearThreadInWasmFlag non_wasm_scope(isolate);
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  Handle<String> str1 = args.at<String>(0);
-  Handle<String> str2 = args.at<String>(1);
+  DirectHandle<String> str1 = args.at<String>(0);
+  DirectHandle<String> str2 = args.at<String>(1);
   RETURN_RESULT_OR_FAILURE(isolate,
                            isolate->factory()->NewConsString(str1, str2));
 }
@@ -204,15 +168,16 @@ RUNTIME_FUNCTION(Runtime_StringAdd) {
 RUNTIME_FUNCTION(Runtime_InternalizeString) {
   HandleScope handles(isolate);
   DCHECK_EQ(1, args.length());
-  Handle<String> string = args.at<String>(0);
+  DirectHandle<String> string = args.at<String>(0);
   return *isolate->factory()->InternalizeString(string);
 }
 
 RUNTIME_FUNCTION(Runtime_StringCharCodeAt) {
+  SaveAndClearThreadInWasmFlag non_wasm_scope(isolate);
   HandleScope handle_scope(isolate);
   DCHECK_EQ(2, args.length());
 
-  Handle<String> subject = args.at<String>(0);
+  DirectHandle<String> subject = args.at<String>(0);
   uint32_t i = NumberToUint32(args[1]);
 
   // Flatten the string.  If someone wants to get a char at an index
@@ -231,7 +196,7 @@ RUNTIME_FUNCTION(Runtime_StringCodePointAt) {
   HandleScope handle_scope(isolate);
   DCHECK_EQ(2, args.length());
 
-  Handle<String> subject = args.at<String>(0);
+  DirectHandle<String> subject = args.at<String>(0);
   uint32_t i = NumberToUint32(args[1]);
 
   // Flatten the string.  If someone wants to get a char at an index
@@ -265,11 +230,11 @@ RUNTIME_FUNCTION(Runtime_StringCodePointAt) {
 RUNTIME_FUNCTION(Runtime_StringBuilderConcat) {
   HandleScope scope(isolate);
   DCHECK_EQ(3, args.length());
-  Handle<FixedArray> array = args.at<FixedArray>(0);
+  DirectHandle<FixedArray> array = args.at<FixedArray>(0);
 
   int array_length = args.smi_value_at(1);
 
-  Handle<String> special = args.at<String>(2);
+  DirectHandle<String> special = args.at<String>(2);
 
   // This assumption is used by the slice encoding in one or two smis.
   DCHECK_GE(Smi::kMaxValue, String::kMaxLength);
@@ -301,7 +266,7 @@ RUNTIME_FUNCTION(Runtime_StringBuilderConcat) {
   }
 
   if (one_byte) {
-    Handle<SeqOneByteString> answer;
+    DirectHandle<SeqOneByteString> answer;
     ASSIGN_RETURN_FAILURE_ON_EXCEPTION(
         isolate, answer, isolate->factory()->NewRawOneByteString(length));
     DisallowGarbageCollection no_gc;
@@ -309,7 +274,7 @@ RUNTIME_FUNCTION(Runtime_StringBuilderConcat) {
                               array_length);
     return *answer;
   } else {
-    Handle<SeqTwoByteString> answer;
+    DirectHandle<SeqTwoByteString> answer;
     ASSIGN_RETURN_FAILURE_ON_EXCEPTION(
         isolate, answer, isolate->factory()->NewRawTwoByteString(length));
     DisallowGarbageCollection no_gc;
@@ -324,20 +289,20 @@ RUNTIME_FUNCTION(Runtime_StringBuilderConcat) {
 RUNTIME_FUNCTION(Runtime_StringToArray) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  Handle<String> s = args.at<String>(0);
+  DirectHandle<String> s = args.at<String>(0);
   uint32_t limit = NumberToUint32(args[1]);
 
   s = String::Flatten(isolate, s);
   const int length =
       static_cast<int>(std::min(static_cast<uint32_t>(s->length()), limit));
 
-  Handle<FixedArray> elements = isolate->factory()->NewFixedArray(length);
+  DirectHandle<FixedArray> elements = isolate->factory()->NewFixedArray(length);
   bool elements_are_initialized = false;
 
   if (s->IsFlat() && s->IsOneByteRepresentation()) {
     DisallowGarbageCollection no_gc;
     String::FlatContent content = s->GetFlatContent(no_gc);
-    // Use pre-initialized single characters to intialize all the elements.
+    // Use pre-initialized single characters to initialize all the elements.
     // This can be false if the string is sliced from an externalized
     // two-byte string that has only one-byte chars, in that case we will do
     // a LookupSingleCharacterStringFromCode for each of the characters.
@@ -348,7 +313,7 @@ RUNTIME_FUNCTION(Runtime_StringToArray) {
       for (int i = 0; i < length; ++i) {
         Tagged<Object> value = one_byte_table->get(chars[i]);
         DCHECK(IsString(value));
-        DCHECK(ReadOnlyHeap::Contains(HeapObject::cast(value)));
+        DCHECK(ReadOnlyHeap::Contains(Cast<HeapObject>(value)));
         // The single-character strings are in RO space so it should
         // be safe to skip the write barriers.
         elements->set(i, value, SKIP_WRITE_BARRIER);
@@ -359,7 +324,7 @@ RUNTIME_FUNCTION(Runtime_StringToArray) {
 
   if (!elements_are_initialized) {
     for (int i = 0; i < length; ++i) {
-      Handle<Object> str =
+      DirectHandle<Object> str =
           isolate->factory()->LookupSingleCharacterStringFromCode(s->Get(i));
       elements->set(i, *str);
     }
@@ -367,7 +332,7 @@ RUNTIME_FUNCTION(Runtime_StringToArray) {
 
 #ifdef DEBUG
   for (int i = 0; i < length; ++i) {
-    DCHECK_EQ(String::cast(elements->get(i))->length(), 1);
+    DCHECK_EQ(Cast<String>(elements->get(i))->length(), 1);
   }
 #endif
 
@@ -377,8 +342,8 @@ RUNTIME_FUNCTION(Runtime_StringToArray) {
 RUNTIME_FUNCTION(Runtime_StringLessThan) {
   HandleScope handle_scope(isolate);
   DCHECK_EQ(2, args.length());
-  Handle<String> x = args.at<String>(0);
-  Handle<String> y = args.at<String>(1);
+  DirectHandle<String> x = args.at<String>(0);
+  DirectHandle<String> y = args.at<String>(1);
   ComparisonResult result = String::Compare(isolate, x, y);
   DCHECK_NE(result, ComparisonResult::kUndefined);
   return isolate->heap()->ToBoolean(
@@ -388,8 +353,8 @@ RUNTIME_FUNCTION(Runtime_StringLessThan) {
 RUNTIME_FUNCTION(Runtime_StringLessThanOrEqual) {
   HandleScope handle_scope(isolate);
   DCHECK_EQ(2, args.length());
-  Handle<String> x = args.at<String>(0);
-  Handle<String> y = args.at<String>(1);
+  DirectHandle<String> x = args.at<String>(0);
+  DirectHandle<String> y = args.at<String>(1);
   ComparisonResult result = String::Compare(isolate, x, y);
   DCHECK_NE(result, ComparisonResult::kUndefined);
   return isolate->heap()->ToBoolean(
@@ -399,8 +364,8 @@ RUNTIME_FUNCTION(Runtime_StringLessThanOrEqual) {
 RUNTIME_FUNCTION(Runtime_StringGreaterThan) {
   HandleScope handle_scope(isolate);
   DCHECK_EQ(2, args.length());
-  Handle<String> x = args.at<String>(0);
-  Handle<String> y = args.at<String>(1);
+  DirectHandle<String> x = args.at<String>(0);
+  DirectHandle<String> y = args.at<String>(1);
   ComparisonResult result = String::Compare(isolate, x, y);
   DCHECK_NE(result, ComparisonResult::kUndefined);
   return isolate->heap()->ToBoolean(
@@ -410,8 +375,8 @@ RUNTIME_FUNCTION(Runtime_StringGreaterThan) {
 RUNTIME_FUNCTION(Runtime_StringGreaterThanOrEqual) {
   HandleScope handle_scope(isolate);
   DCHECK_EQ(2, args.length());
-  Handle<String> x = args.at<String>(0);
-  Handle<String> y = args.at<String>(1);
+  DirectHandle<String> x = args.at<String>(0);
+  DirectHandle<String> y = args.at<String>(1);
   ComparisonResult result = String::Compare(isolate, x, y);
   DCHECK_NE(result, ComparisonResult::kUndefined);
   return isolate->heap()->ToBoolean(
@@ -419,19 +384,37 @@ RUNTIME_FUNCTION(Runtime_StringGreaterThanOrEqual) {
 }
 
 RUNTIME_FUNCTION(Runtime_StringEqual) {
+  SaveAndClearThreadInWasmFlag non_wasm_scope(isolate);
   HandleScope handle_scope(isolate);
   DCHECK_EQ(2, args.length());
-  Handle<String> x = args.at<String>(0);
-  Handle<String> y = args.at<String>(1);
+#ifdef V8_ENABLE_CONSERVATIVE_STACK_SCANNING
+  DirectHandle<String> x = args.at<String>(0);
+  DirectHandle<String> y = args.at<String>(1);
+#else
+  // This function can be called from Wasm: optimized Wasm code calls
+  // straight to the "StringEqual" builtin, which tail-calls here. So on
+  // the stack, the CEntryStub's EXIT frame will sit right on top of the
+  // Wasm frame; and Wasm frames don't scan their outgoing parameters (in
+  // order to support tail-calls between Wasm functions), while the EXIT
+  // frame doesn't scan its incoming parameters (because it expects to be
+  // called from JS).
+  // Working around this by calling through a trampoline builtin is slow.
+  // Teaching the stack walker to be smarter has proven to be difficult.
+  // In the future, Conservative Stack Scanning will trivially solve the
+  // problem. In the meantime, we can work around it by explicitly creating
+  // handles here (rather than treating the on-stack arguments as handles).
+  DirectHandle<String> x(*args.at<String>(0), isolate);
+  DirectHandle<String> y(*args.at<String>(1), isolate);
+#endif  // V8_ENABLE_CONSERVATIVE_STACK_SCANNING
   return isolate->heap()->ToBoolean(String::Equals(isolate, x, y));
 }
 
 RUNTIME_FUNCTION(Runtime_StringCompare) {
-  CLEAR_THREAD_IN_WASM_SCOPE;
+  SaveAndClearThreadInWasmFlag non_wasm_scope(isolate);
   DCHECK_EQ(2, args.length());
   HandleScope scope(isolate);
-  Handle<String> lhs(String::cast(args[0]), isolate);
-  Handle<String> rhs(String::cast(args[1]), isolate);
+  DirectHandle<String> lhs(Cast<String>(args[0]), isolate);
+  DirectHandle<String> rhs(Cast<String>(args[1]), isolate);
   ComparisonResult result = String::Compare(isolate, lhs, rhs);
   DCHECK_NE(result, ComparisonResult::kUndefined);
   return Smi::FromInt(static_cast<int>(result));
@@ -440,7 +423,7 @@ RUNTIME_FUNCTION(Runtime_StringCompare) {
 RUNTIME_FUNCTION(Runtime_FlattenString) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  Handle<String> str = args.at<String>(0);
+  DirectHandle<String> str = args.at<String>(0);
   return *String::Flatten(isolate, str);
 }
 
@@ -452,13 +435,13 @@ RUNTIME_FUNCTION(Runtime_StringMaxLength) {
 RUNTIME_FUNCTION(Runtime_StringEscapeQuotes) {
   HandleScope handle_scope(isolate);
   DCHECK_EQ(1, args.length());
-  Handle<String> string = args.at<String>(0);
+  DirectHandle<String> string = args.at<String>(0);
 
   // Equivalent to global replacement `string.replace(/"/g, "&quot")`, but this
   // does not modify any global state (e.g. the regexp match info).
 
   const int string_length = string->length();
-  Handle<String> quotes =
+  DirectHandle<String> quotes =
       isolate->factory()->LookupSingleCharacterStringFromCode('"');
 
   int quote_index = String::IndexOf(isolate, string, quotes, 0);
@@ -475,7 +458,7 @@ RUNTIME_FUNCTION(Runtime_StringEscapeQuotes) {
   }
 
   // Build the replacement string.
-  Handle<String> replacement =
+  DirectHandle<String> replacement =
       isolate->factory()->NewStringFromAsciiChecked("&quot;");
   const int estimated_part_count = static_cast<int>(indices.size()) * 2 + 1;
   ReplacementStringBuilder builder(isolate->heap(), string,
@@ -496,13 +479,15 @@ RUNTIME_FUNCTION(Runtime_StringEscapeQuotes) {
     builder.AddSubjectSlice(prev_index + 1, string_length);
   }
 
-  return *builder.ToString().ToHandleChecked();
+  DirectHandle<String> result;
+  ASSIGN_RETURN_FAILURE_ON_EXCEPTION(isolate, result, builder.ToString());
+  return *result;
 }
 
 RUNTIME_FUNCTION(Runtime_StringIsWellFormed) {
   HandleScope handle_scope(isolate);
   DCHECK_EQ(1, args.length());
-  Handle<String> string = args.at<String>(0);
+  DirectHandle<String> string = args.at<String>(0);
   return isolate->heap()->ToBoolean(
       String::IsWellFormedUnicode(isolate, string));
 }
@@ -510,12 +495,12 @@ RUNTIME_FUNCTION(Runtime_StringIsWellFormed) {
 RUNTIME_FUNCTION(Runtime_StringToWellFormed) {
   HandleScope handle_scope(isolate);
   DCHECK_EQ(1, args.length());
-  Handle<String> source = args.at<String>(0);
+  DirectHandle<String> source = args.at<String>(0);
   if (String::IsWellFormedUnicode(isolate, source)) return *source;
   // String::IsWellFormedUnicode would have returned true above otherwise.
   DCHECK(!String::IsOneByteRepresentationUnderneath(*source));
   const int length = source->length();
-  Handle<SeqTwoByteString> dest =
+  DirectHandle<SeqTwoByteString> dest =
       isolate->factory()->NewRawTwoByteString(length).ToHandleChecked();
   DisallowGarbageCollection no_gc;
   String::FlatContent source_contents = source->GetFlatContent(no_gc);
