@@ -6,9 +6,9 @@
 
 #include <cmath>
 #include <iostream>
+#include <optional>
 
 #include "src/base/bits.h"
-#include "src/base/optional.h"
 #include "src/torque/ast.h"
 #include "src/torque/declarable.h"
 #include "src/torque/global-context.h"
@@ -16,9 +16,7 @@
 #include "src/torque/type-oracle.h"
 #include "src/torque/type-visitor.h"
 
-namespace v8 {
-namespace internal {
-namespace torque {
+namespace v8::internal::torque {
 
 // This custom copy constructor doesn't copy aliases_ and id_ because they
 // should be distinct for each type.
@@ -123,31 +121,31 @@ std::string Type::GetConstexprGeneratedTypeName() const {
   return constexpr_version->GetGeneratedTypeName();
 }
 
-base::Optional<const ClassType*> Type::ClassSupertype() const {
+std::optional<const ClassType*> Type::ClassSupertype() const {
   for (const Type* t = this; t != nullptr; t = t->parent()) {
     if (auto* class_type = ClassType::DynamicCast(t)) {
       return class_type;
     }
   }
-  return base::nullopt;
+  return std::nullopt;
 }
 
-base::Optional<const StructType*> Type::StructSupertype() const {
+std::optional<const StructType*> Type::StructSupertype() const {
   for (const Type* t = this; t != nullptr; t = t->parent()) {
     if (auto* struct_type = StructType::DynamicCast(t)) {
       return struct_type;
     }
   }
-  return base::nullopt;
+  return std::nullopt;
 }
 
-base::Optional<const AggregateType*> Type::AggregateSupertype() const {
+std::optional<const AggregateType*> Type::AggregateSupertype() const {
   for (const Type* t = this; t != nullptr; t = t->parent()) {
     if (auto* aggregate_type = AggregateType::DynamicCast(t)) {
       return aggregate_type;
     }
   }
-  return base::nullopt;
+  return std::nullopt;
 }
 
 // static
@@ -201,7 +199,7 @@ std::string AbstractType::GetGeneratedTypeNameImpl() const {
   // A special case that is not very well represented by the "generates"
   // syntax in the .tq files: Lazy<T> represents a std::function that
   // produces a TNode of the wrapped type.
-  if (base::Optional<const Type*> type_wrapped_in_lazy =
+  if (std::optional<const Type*> type_wrapped_in_lazy =
           Type::MatchUnaryGeneric(this, TypeOracle::GetLazyGeneric())) {
     DCHECK(!IsConstexpr());
     return "std::function<" + (*type_wrapped_in_lazy)->GetGeneratedTypeName() +
@@ -283,22 +281,109 @@ std::string UnionType::SimpleNameImpl() const {
   return result.str();
 }
 
+// static
+void UnionType::InsertGeneratedTNodeTypeName(std::set<std::string>& names,
+                                             const Type* t) {
+  if (t->IsUnionType()) {
+    for (const Type* u : ((const UnionType*)t)->types_) {
+      names.insert(u->GetGeneratedTNodeTypeName());
+    }
+  } else {
+    names.insert(t->GetGeneratedTNodeTypeName());
+  }
+}
+
 std::string UnionType::GetGeneratedTNodeTypeNameImpl() const {
-  if (types_.size() <= 3) {
-    std::set<std::string> members;
-    for (const Type* t : types_) {
-      members.insert(t->GetGeneratedTNodeTypeName());
-    }
-    if (members == std::set<std::string>{"Smi", "HeapNumber"}) {
-      return "Number";
-    }
-    if (members == std::set<std::string>{"Smi", "HeapNumber", "BigInt"}) {
-      return "Numeric";
+  // For non-tagged unions, use the parent GetGeneratedTNodeTypeName.
+  for (const Type* t : types_) {
+    if (!t->IsSubtypeOf(TypeOracle::GetTaggedType())) {
+      return parent()->GetGeneratedTNodeTypeName();
     }
   }
-  return parent()->GetGeneratedTNodeTypeName();
+
+  std::string simple_name = SimpleName();
+  if (simple_name == "Object") return simple_name;
+  if (simple_name == "Number") return simple_name;
+  if (simple_name == "Numeric") return simple_name;
+  if (simple_name == "JSAny") return simple_name;
+  if (simple_name == "JSPrimitive") return simple_name;
+
+  std::set<std::string> names;
+  for (const Type* t : types_) {
+    InsertGeneratedTNodeTypeName(names, t);
+  }
+  std::stringstream result;
+  result << "Union<";
+  bool first = true;
+  for (std::string name : names) {
+    if (!first) {
+      result << ", ";
+    }
+    first = false;
+    result << name;
+  }
+  result << ">";
+  return result.str();
 }
 
+std::string UnionType::GetRuntimeType() const {
+  for (const Type* t : types_) {
+    if (!t->IsSubtypeOf(TypeOracle::GetTaggedType())) {
+      return parent()->GetRuntimeType();
+    }
+  }
+  return "Tagged<" + GetConstexprGeneratedTypeName() + ">";
+}
+
+// static
+void UnionType::InsertConstexprGeneratedTypeName(std::set<std::string>& names,
+                                                 const Type* t) {
+  if (t->IsUnionType()) {
+    for (const Type* u : ((const UnionType*)t)->types_) {
+      names.insert(u->GetConstexprGeneratedTypeName());
+    }
+  } else {
+    names.insert(t->GetConstexprGeneratedTypeName());
+  }
+}
+
+std::string UnionType::GetConstexprGeneratedTypeName() const {
+  // For non-tagged unions, use the superclass GetConstexprGeneratedTypeName.
+  for (const Type* t : types_) {
+    if (!t->IsSubtypeOf(TypeOracle::GetTaggedType())) {
+      return this->Type::GetConstexprGeneratedTypeName();
+    }
+  }
+
+  // Allow some aliased simple names to be used as-is.
+  std::string simple_name = SimpleName();
+  if (simple_name == "Object") return simple_name;
+  if (simple_name == "Number") return simple_name;
+  if (simple_name == "Numeric") return simple_name;
+  if (simple_name == "JSAny") return simple_name;
+  if (simple_name == "JSPrimitive") return simple_name;
+
+  // Deduplicate generated typenames and flatten unions.
+  std::set<std::string> names;
+  for (const Type* t : types_) {
+    InsertConstexprGeneratedTypeName(names, t);
+  }
+  std::stringstream result;
+  result << "Union<";
+  bool first = true;
+  for (std::string name : names) {
+    if (!first) {
+      result << ", ";
+    }
+    first = false;
+    result << name;
+  }
+  result << ">";
+  return result.str();
+}
+
+std::string UnionType::GetDebugType() const { return parent()->GetDebugType(); }
+
 void UnionType::RecomputeParent() {
   const Type* parent = nullptr;
   for (const Type* t : types_) {
@@ -485,15 +570,15 @@ std::string Type::ComputeName(const std::string& basename,
 std::string StructType::SimpleNameImpl() const { return decl_->name->value; }
 
 // static
-base::Optional<const Type*> Type::MatchUnaryGeneric(const Type* type,
-                                                    GenericType* generic) {
+std::optional<const Type*> Type::MatchUnaryGeneric(const Type* type,
+                                                   GenericType* generic) {
   DCHECK_EQ(generic->generic_parameters().size(), 1);
   if (!type->GetSpecializedFrom()) {
-    return base::nullopt;
+    return std::nullopt;
   }
   auto& key = type->GetSpecializedFrom().value();
   if (key.generic != generic || key.specialized_types.size() != 1) {
-    return base::nullopt;
+    return std::nullopt;
   }
   return {key.specialized_types[0]};
 }
@@ -598,19 +683,19 @@ std::vector<Field> ClassType::ComputeArrayFields() const {
 }
 
 void ClassType::InitializeInstanceTypes(
-    base::Optional<int> own, base::Optional<std::pair<int, int>> range) const {
+    std::optional<int> own, std::optional<std::pair<int, int>> range) const {
   DCHECK(!own_instance_type_.has_value());
   DCHECK(!instance_type_range_.has_value());
   own_instance_type_ = own;
   instance_type_range_ = range;
 }
 
-base::Optional<int> ClassType::OwnInstanceType() const {
+std::optional<int> ClassType::OwnInstanceType() const {
   DCHECK(GlobalContext::IsInstanceTypesInitialized());
   return own_instance_type_;
 }
 
-base::Optional<std::pair<int, int>> ClassType::InstanceTypeRange() const {
+std::optional<std::pair<int, int>> ClassType::InstanceTypeRange() const {
   DCHECK(GlobalContext::IsInstanceTypesInitialized());
   return instance_type_range_;
 }
@@ -669,10 +754,10 @@ std::vector<ObjectSlotKind> ClassType::ComputeHeaderSlotKinds() const {
   return result;
 }
 
-base::Optional<ObjectSlotKind> ClassType::ComputeArraySlotKind() const {
+std::optional<ObjectSlotKind> ClassType::ComputeArraySlotKind() const {
   std::vector<ObjectSlotKind> kinds;
   ComputeSlotKindsHelper(&kinds, 0, ComputeArrayFields());
-  if (kinds.empty()) return base::nullopt;
+  if (kinds.empty()) return std::nullopt;
   std::sort(kinds.begin(), kinds.end());
   if (kinds.front() == kinds.back()) return {kinds.front()};
   if (kinds.front() == ObjectSlotKind::kStrongPointer &&
@@ -792,8 +877,8 @@ void ClassType::GenerateAccessors() {
           MakeNode<ElementAccessExpression>(load_expression, index);
     }
     Statement* load_body = MakeNode<ReturnStatement>(load_expression);
-    Declarations::DeclareMacro(load_macro_name, true, base::nullopt,
-                               load_signature, load_body, base::nullopt);
+    Declarations::DeclareMacro(load_macro_name, true, std::nullopt,
+                               load_signature, load_body, std::nullopt);
 
     // Store accessor
     if (!field.const_qualified) {
@@ -820,8 +905,8 @@ void ClassType::GenerateAccessors() {
       }
       Statement* store_body = MakeNode<ExpressionStatement>(
           MakeNode<AssignmentExpression>(store_expression, value));
-      Declarations::DeclareMacro(store_macro_name, true, base::nullopt,
-                                 store_signature, store_body, base::nullopt,
+      Declarations::DeclareMacro(store_macro_name, true, std::nullopt,
+                                 store_signature, store_body, std::nullopt,
                                  false);
     }
   }
@@ -948,8 +1033,8 @@ void ClassType::GenerateSliceAccessor(size_t field_index) {
   Statement* block =
       MakeNode<BlockStatement>(/*deferred=*/false, std::move(statements));
 
-  Macro* macro = Declarations::DeclareMacro(macro_name, true, base::nullopt,
-                                            signature, block, base::nullopt);
+  Macro* macro = Declarations::DeclareMacro(macro_name, true, std::nullopt,
+                                            signature, block, std::nullopt);
   if (this->ShouldGenerateCppObjectLayoutDefinitionAsserts()) {
     GlobalContext::EnsureInCCDebugOutputList(TorqueMacro::cast(macro),
                                              macro->Position().source);
@@ -1066,7 +1151,7 @@ bool Signature::HasSameTypesAs(const Signature& other,
 }
 
 namespace {
-bool FirstTypeIsContext(const std::vector<const Type*> parameter_types) {
+bool FirstTypeIsContext(const std::vector<const Type*>& parameter_types) {
   return !parameter_types.empty() &&
          (parameter_types[0] == TypeOracle::GetContextType() ||
           parameter_types[0] == TypeOracle::GetNoContextType());
@@ -1112,7 +1197,7 @@ namespace {
 void AppendLoweredTypes(const Type* type, std::vector<const Type*>* result) {
   if (type->IsConstexpr()) return;
   if (type->IsVoidOrNever()) return;
-  if (base::Optional<const StructType*> s = type->StructSupertype()) {
+  if (std::optional<const StructType*> s = type->StructSupertype()) {
     for (const Field& field : (*s)->fields()) {
       AppendLoweredTypes(field.name_and_type.type, result);
     }
@@ -1184,8 +1269,10 @@ size_t AbstractType::AlignmentLog2() const {
     alignment = TargetArchitecture::RawPtrSize();
   } else if (this == TypeOracle::GetExternalPointerType()) {
     alignment = TargetArchitecture::ExternalPointerSize();
-  } else if (this == TypeOracle::GetIndirectPointerType()) {
-    alignment = TargetArchitecture::IndirectPointerSize();
+  } else if (this == TypeOracle::GetCppHeapPointerType()) {
+    alignment = TargetArchitecture::CppHeapPointerSize();
+  } else if (this == TypeOracle::GetTrustedPointerType()) {
+    alignment = TargetArchitecture::TrustedPointerSize();
   } else if (this == TypeOracle::GetProtectedPointerType()) {
     alignment = TargetArchitecture::ProtectedPointerSize();
   } else if (this == TypeOracle::GetVoidType()) {
@@ -1216,7 +1303,7 @@ size_t AbstractType::AlignmentLog2() const {
 }
 
 size_t StructType::AlignmentLog2() const {
-  if (this == TypeOracle::GetFloat64OrHoleType()) {
+  if (this == TypeOracle::GetFloat64OrUndefinedOrHoleType()) {
     return TypeOracle::GetFloat64Type()->AlignmentLog2();
   }
   size_t alignment_log_2 = 0;
@@ -1229,8 +1316,9 @@ size_t StructType::AlignmentLog2() const {
 
 void Field::ValidateAlignment(ResidueClass at_offset) const {
   const Type* type = name_and_type.type;
-  base::Optional<const StructType*> struct_type = type->StructSupertype();
-  if (struct_type && struct_type != TypeOracle::GetFloat64OrHoleType()) {
+  std::optional<const StructType*> struct_type = type->StructSupertype();
+  if (struct_type &&
+      struct_type != TypeOracle::GetFloat64OrUndefinedOrHoleType()) {
     for (const Field& field : (*struct_type)->fields()) {
       field.ValidateAlignment(at_offset);
       size_t field_size = std::get<0>(field.GetFieldSizeInformation());
@@ -1246,7 +1334,7 @@ void Field::ValidateAlignment(ResidueClass at_offset) const {
   }
 }
 
-base::Optional<std::tuple<size_t, std::string>> SizeOf(const Type* type) {
+std::optional<std::tuple<size_t, std::string>> SizeOf(const Type* type) {
   std::string size_string;
   size_t size;
   if (type->IsSubtypeOf(TypeOracle::GetTaggedType())) {
@@ -1258,9 +1346,12 @@ base::Optional<std::tuple<size_t, std::string>> SizeOf(const Type* type) {
   } else if (type->IsSubtypeOf(TypeOracle::GetExternalPointerType())) {
     size = TargetArchitecture::ExternalPointerSize();
     size_string = "kExternalPointerSlotSize";
-  } else if (type->IsSubtypeOf(TypeOracle::GetIndirectPointerType())) {
-    size = TargetArchitecture::IndirectPointerSize();
-    size_string = "kIndirectPointerSize";
+  } else if (type->IsSubtypeOf(TypeOracle::GetCppHeapPointerType())) {
+    size = TargetArchitecture::CppHeapPointerSize();
+    size_string = "kCppHeapPointerSlotSize";
+  } else if (type->IsSubtypeOf(TypeOracle::GetTrustedPointerType())) {
+    size = TargetArchitecture::TrustedPointerSize();
+    size_string = "kTrustedPointerSize";
   } else if (type->IsSubtypeOf(TypeOracle::GetProtectedPointerType())) {
     size = TargetArchitecture::ProtectedPointerSize();
     size_string = "kTaggedSize";
@@ -1295,7 +1386,7 @@ base::Optional<std::tuple<size_t, std::string>> SizeOf(const Type* type) {
     size = TargetArchitecture::RawPtrSize();
     size_string = "kIntptrSize";
   } else if (auto struct_type = type->StructSupertype()) {
-    if (type == TypeOracle::GetFloat64OrHoleType()) {
+    if (type == TypeOracle::GetFloat64OrUndefinedOrHoleType()) {
       size = kDoubleSize;
       size_string = "kDoubleSize";
     } else {
@@ -1339,7 +1430,7 @@ bool Is32BitIntegralType(const Type* type) {
          type->IsSubtypeOf(TypeOracle::GetBoolType());
 }
 
-base::Optional<NameAndType> ExtractSimpleFieldArraySize(
+std::optional<NameAndType> ExtractSimpleFieldArraySize(
     const ClassType& class_type, Expression* array_size) {
   IdentifierExpression* identifier =
       IdentifierExpression::DynamicCast(array_size);
@@ -1355,7 +1446,7 @@ std::string Type::GetRuntimeType() const {
   if (IsSubtypeOf(TypeOracle::GetTaggedType())) {
     return "Tagged<" + GetGeneratedTNodeTypeName() + ">";
   }
-  if (base::Optional<const StructType*> struct_type = StructSupertype()) {
+  if (std::optional<const StructType*> struct_type = StructSupertype()) {
     std::stringstream result;
     result << "std::tuple<";
     bool first = true;
@@ -1375,7 +1466,7 @@ std::string Type::GetDebugType() const {
   if (IsSubtypeOf(TypeOracle::GetTaggedType())) {
     return "uintptr_t";
   }
-  if (base::Optional<const StructType*> struct_type = StructSupertype()) {
+  if (std::optional<const StructType*> struct_type = StructSupertype()) {
     std::stringstream result;
     result << "std::tuple<";
     bool first = true;
@@ -1390,6 +1481,4 @@ std::string Type::GetDebugType() const {
   return ConstexprVersion()->GetGeneratedTypeName();
 }
 
-}  // namespace torque
-}  // namespace internal
-}  // namespace v8
+}  // namespace v8::internal::torque
