@@ -5,7 +5,7 @@
 #ifndef V8_WASM_BASELINE_RISCV_LIFTOFF_ASSEMBLER_RISCV32_INL_H_
 #define V8_WASM_BASELINE_RISCV_LIFTOFF_ASSEMBLER_RISCV32_INL_H_
 
-#include "src/heap/mutable-page.h"
+#include "src/heap/mutable-page-metadata.h"
 #include "src/wasm/baseline/liftoff-assembler.h"
 #include "src/wasm/baseline/riscv/liftoff-assembler-riscv-inl.h"
 #include "src/wasm/wasm-objects.h"
@@ -90,7 +90,6 @@ inline void Load(LiftoffAssembler* assm, LiftoffRegister dst, Register base,
     case kI32:
     case kRef:
     case kRefNull:
-    case kRtt:
       assm->Lw(dst.gp(), src);
       break;
     case kI64:
@@ -126,7 +125,6 @@ inline void Store(LiftoffAssembler* assm, Register base, int32_t offset,
     case kI32:
     case kRefNull:
     case kRef:
-    case kRtt:
       assm->Sw(src.gp(), dst);
       break;
     case kI64:
@@ -160,7 +158,6 @@ inline void push(LiftoffAssembler* assm, LiftoffRegister reg, ValueKind kind) {
     case kI32:
     case kRefNull:
     case kRef:
-    case kRtt:
       assm->addi(sp, sp, -kSystemPointerSize);
       assm->Sw(reg.gp(), MemOperand(sp, 0));
       break;
@@ -224,21 +221,6 @@ void LiftoffAssembler::LoadConstant(LiftoffRegister reg, WasmValue value) {
   }
 }
 
-void LiftoffAssembler::LoadExternalPointer(Register dst, Register src_addr,
-                                           int offset, ExternalPointerTag tag,
-                                           Register scratch) {
-  LoadFullPointer(dst, src_addr, offset);
-}
-
-void LiftoffAssembler::LoadExternalPointer(Register dst, Register src_addr,
-                                           int offset, Register index,
-                                           ExternalPointerTag tag,
-                                           Register scratch) {
-  MemOperand src_op =
-      liftoff::GetMemOp(this, src_addr, index, offset, kSystemPointerSizeLog2);
-  LoadWord(dst, src_op);
-}
-
 void LiftoffAssembler::LoadTaggedPointer(Register dst, Register src_addr,
                                          Register offset_reg,
                                          int32_t offset_imm,
@@ -288,10 +270,13 @@ void LiftoffAssembler::StoreTaggedPointer(Register dst_addr,
     AddWord(kScratchReg, dst_addr, actual_offset_reg);
     dst_op = MemOperand(kScratchReg, 0);
   }
-
-  if (protected_store_pc) *protected_store_pc = pc_offset();
-
-  StoreWord(src, dst_op);
+  auto trapper = [protected_store_pc](int offset) {
+    if (protected_store_pc) *protected_store_pc = static_cast<uint32_t>(offset);
+  };
+  StoreWord(src, dst_op, trapper);
+  if (protected_store_pc) {
+    DCHECK(InstructionAt(*protected_store_pc)->IsStore());
+  }
 
   if (skip_write_barrier || v8_flags.disable_write_barriers) return;
 
@@ -317,72 +302,80 @@ void LiftoffAssembler::Load(LiftoffRegister dst, Register src_addr,
   unsigned shift_amount = needs_shift ? type.size_log_2() : 0;
   MemOperand src_op =
       liftoff::GetMemOp(this, src_addr, offset_reg, offset_imm, shift_amount);
-
-  if (protected_load_pc) *protected_load_pc = pc_offset();
+  auto trapper = [protected_load_pc](int offset) {
+    if (protected_load_pc) *protected_load_pc = static_cast<uint32_t>(offset);
+  };
   switch (type.value()) {
     case LoadType::kI32Load8U:
-      Lbu(dst.gp(), src_op);
+      Lbu(dst.gp(), src_op, trapper);
       break;
     case LoadType::kI64Load8U:
-      Lbu(dst.low_gp(), src_op);
-      MacroAssembler::mv(dst.high_gp(), zero_reg);
+      Lbu(dst.low_gp(), src_op, trapper);
+      mv(dst.high_gp(), zero_reg);
       break;
     case LoadType::kI32Load8S:
-      Lb(dst.gp(), src_op);
+      Lb(dst.gp(), src_op, trapper);
       break;
     case LoadType::kI64Load8S:
-      Lb(dst.low_gp(), src_op);
-      MacroAssembler::srai(dst.high_gp(), dst.low_gp(), 31);
+      Lb(dst.low_gp(), src_op, trapper);
+      srai(dst.high_gp(), dst.low_gp(), 31);
       break;
     case LoadType::kI32Load16U:
-      MacroAssembler::Lhu(dst.gp(), src_op);
+      Lhu(dst.gp(), src_op, trapper);
       break;
     case LoadType::kI64Load16U:
-      MacroAssembler::Lhu(dst.low_gp(), src_op);
-      MacroAssembler::mv(dst.high_gp(), zero_reg);
+      Lhu(dst.low_gp(), src_op, trapper);
+      mv(dst.high_gp(), zero_reg);
       break;
     case LoadType::kI32Load16S:
-      MacroAssembler::Lh(dst.gp(), src_op);
+      Lh(dst.gp(), src_op, trapper);
       break;
     case LoadType::kI64Load16S:
-      MacroAssembler::Lh(dst.low_gp(), src_op);
-      MacroAssembler::srai(dst.high_gp(), dst.low_gp(), 31);
+      Lh(dst.low_gp(), src_op, trapper);
+      srai(dst.high_gp(), dst.low_gp(), 31);
       break;
     case LoadType::kI64Load32U:
-      MacroAssembler::Lw(dst.low_gp(), src_op);
-      MacroAssembler::mv(dst.high_gp(), zero_reg);
+      Lw(dst.low_gp(), src_op, trapper);
+      mv(dst.high_gp(), zero_reg);
       break;
     case LoadType::kI64Load32S:
-      MacroAssembler::Lw(dst.low_gp(), src_op);
-      MacroAssembler::srai(dst.high_gp(), dst.low_gp(), 31);
+      Lw(dst.low_gp(), src_op, trapper);
+      srai(dst.high_gp(), dst.low_gp(), 31);
       break;
     case LoadType::kI32Load:
-      MacroAssembler::Lw(dst.gp(), src_op);
+      Lw(dst.gp(), src_op, trapper);
       break;
     case LoadType::kI64Load: {
-      Lw(dst.low_gp(), src_op);
+      Lw(dst.low_gp(), src_op, trapper);
       src_op = liftoff::GetMemOp(this, src_addr, offset_reg,
                                  offset_imm + kSystemPointerSize);
       Lw(dst.high_gp(), src_op);
     } break;
     case LoadType::kF32Load:
-      MacroAssembler::LoadFloat(dst.fp(), src_op);
+      LoadFloat(dst.fp(), src_op, trapper);
       break;
     case LoadType::kF64Load:
-      MacroAssembler::LoadDouble(dst.fp(), src_op);
+      LoadDouble(dst.fp(), src_op, trapper);
       break;
     case LoadType::kS128Load: {
       VU.set(kScratchReg, E8, m1);
       Register src_reg = src_op.offset() == 0 ? src_op.rm() : kScratchReg;
       if (src_op.offset() != 0) {
-        MacroAssembler::AddWord(src_reg, src_op.rm(), src_op.offset());
+        AddWord(src_reg, src_op.rm(), src_op.offset());
       }
+      trapper(pc_offset());
       vl(dst.fp().toV(), src_reg, 0, E8);
       break;
     }
+    case LoadType::kF32LoadF16:
+      UNIMPLEMENTED();
+      break;
     default:
       UNREACHABLE();
   }
+  if (protected_load_pc) {
+    DCHECK(InstructionAt(*protected_load_pc)->IsLoad());
+  }
 
 #if defined(V8_TARGET_BIG_ENDIAN)
   if (is_load_mem) {
@@ -411,40 +404,40 @@ void LiftoffAssembler::Store(Register dst_addr, Register offset_reg,
     liftoff::ChangeEndiannessStore(this, src, type, pinned);
   }
 #endif
-
-  if (protected_store_pc) *protected_store_pc = pc_offset();
-
+  auto trapper = [protected_store_pc](int offset) {
+    if (protected_store_pc) *protected_store_pc = static_cast<uint32_t>(offset);
+  };
   switch (type.value()) {
     case StoreType::kI32Store8:
-      Sb(src.gp(), dst_op);
+      Sb(src.gp(), dst_op, trapper);
       break;
     case StoreType::kI64Store8:
-      Sb(src.low_gp(), dst_op);
+      Sb(src.low_gp(), dst_op, trapper);
       break;
     case StoreType::kI32Store16:
-      MacroAssembler::Sh(src.gp(), dst_op);
+      Sh(src.gp(), dst_op, trapper);
       break;
     case StoreType::kI64Store16:
-      MacroAssembler::Sh(src.low_gp(), dst_op);
+      Sh(src.low_gp(), dst_op, trapper);
       break;
     case StoreType::kI32Store:
-      MacroAssembler::Sw(src.gp(), dst_op);
+      Sw(src.gp(), dst_op, trapper);
       break;
     case StoreType::kI64Store32:
-      MacroAssembler::Sw(src.low_gp(), dst_op);
+      Sw(src.low_gp(), dst_op, trapper);
       break;
     case StoreType::kI64Store: {
-      MacroAssembler::Sw(src.low_gp(), dst_op);
+      Sw(src.low_gp(), dst_op, trapper);
       dst_op = liftoff::GetMemOp(this, dst_addr, offset_reg,
                                  offset_imm + kSystemPointerSize);
-      MacroAssembler::Sw(src.high_gp(), dst_op);
+      Sw(src.high_gp(), dst_op, trapper);
       break;
     }
     case StoreType::kF32Store:
-      MacroAssembler::StoreFloat(src.fp(), dst_op);
+      StoreFloat(src.fp(), dst_op, trapper);
       break;
     case StoreType::kF64Store:
-      MacroAssembler::StoreDouble(src.fp(), dst_op);
+      StoreDouble(src.fp(), dst_op, trapper);
       break;
     case StoreType::kS128Store: {
       VU.set(kScratchReg, E8, m1);
@@ -452,12 +445,16 @@ void LiftoffAssembler::Store(Register dst_addr, Register offset_reg,
       if (dst_op.offset() != 0) {
         AddWord(kScratchReg, dst_op.rm(), dst_op.offset());
       }
+      trapper(pc_offset());
       vs(src.fp().toV(), dst_reg, 0, VSew::E8);
       break;
     }
     default:
       UNREACHABLE();
   }
+  if (protected_store_pc) {
+    DCHECK(InstructionAt(*protected_store_pc)->IsStore());
+  }
 }
 
 namespace liftoff {
@@ -1038,9 +1035,10 @@ void LiftoffAssembler::LoadCallerFrameSlot(LiftoffRegister dst,
 
 void LiftoffAssembler::StoreCallerFrameSlot(LiftoffRegister src,
                                             uint32_t caller_slot_idx,
-                                            ValueKind kind) {
+                                            ValueKind kind,
+                                            Register frame_pointer) {
   int32_t offset = kSystemPointerSize * (caller_slot_idx + 1);
-  liftoff::Store(this, fp, offset, src, kind);
+  liftoff::Store(this, frame_pointer, offset, src, kind);
 }
 
 void LiftoffAssembler::LoadReturnStackSlot(LiftoffRegister dst, int offset,
@@ -1062,7 +1060,6 @@ void LiftoffAssembler::MoveStackValue(uint32_t dst_offset, uint32_t src_offset,
     case kI64:
     case kRef:
     case kRefNull:
-    case kRtt:
       Lw(kScratchReg, src);
       Sw(kScratchReg, dst);
       src = liftoff::GetStackSlot(src_offset - 4);
@@ -1121,7 +1118,6 @@ void LiftoffAssembler::Spill(int offset, LiftoffRegister reg, ValueKind kind) {
     case kI32:
     case kRef:
     case kRefNull:
-    case kRtt:
       Sw(reg.gp(), dst);
       break;
     case kI64:
@@ -1881,18 +1877,20 @@ void LiftoffAssembler::LoadTransform(LiftoffRegister dst, Register src_addr,
                                      Register offset_reg, uintptr_t offset_imm,
                                      LoadType type,
                                      LoadTransformationKind transform,
-                                     uint32_t* protected_load_pc) {
+                                     uint32_t* protected_load_pc,
+                                     bool i64_offset) {
   UseScratchRegisterScope temps(this);
   Register scratch = temps.Acquire();
   MemOperand src_op = liftoff::GetMemOp(this, src_addr, offset_reg, offset_imm);
   VRegister dst_v = dst.fp().toV();
-  *protected_load_pc = pc_offset();
-
+  auto trapper = [protected_load_pc](int offset) {
+    if (protected_load_pc) *protected_load_pc = static_cast<uint32_t>(offset);
+  };
   MachineType memtype = type.mem_type();
   if (transform == LoadTransformationKind::kExtend) {
     // TODO(RISCV): need to confirm the performance impact of using floating
     // point registers.
-    LoadDouble(kScratchDoubleReg, src_op);
+    LoadDouble(kScratchDoubleReg, src_op, trapper);
     if (memtype == MachineType::Int8()) {
       VU.set(kScratchReg, E64, m1);
       vfmv_vf(kSimd128ScratchReg, kScratchDoubleReg);
@@ -1926,67 +1924,72 @@ void LiftoffAssembler::LoadTransform(LiftoffRegister dst, Register src_addr,
     vxor_vv(dst_v, dst_v, dst_v);
     if (memtype == MachineType::Int32()) {
       VU.set(kScratchReg, E32, m1);
-      Lw(scratch, src_op);
+      Lw(scratch, src_op, trapper);
       vmv_sx(dst_v, scratch);
     } else {
       DCHECK_EQ(MachineType::Int64(), memtype);
       VU.set(kScratchReg, E64, m1);
-      LoadDouble(kScratchDoubleReg, src_op);
+      LoadDouble(kScratchDoubleReg, src_op, trapper);
       vfmv_sf(dst_v, kScratchDoubleReg);
     }
   } else {
     DCHECK_EQ(LoadTransformationKind::kSplat, transform);
     if (memtype == MachineType::Int8()) {
       VU.set(kScratchReg, E8, m1);
-      Lb(scratch, src_op);
+      Lb(scratch, src_op, trapper);
       vmv_vx(dst_v, scratch);
     } else if (memtype == MachineType::Int16()) {
       VU.set(kScratchReg, E16, m1);
-      Lh(scratch, src_op);
+      Lh(scratch, src_op, trapper);
       vmv_vx(dst_v, scratch);
     } else if (memtype == MachineType::Int32()) {
       VU.set(kScratchReg, E32, m1);
-      Lw(scratch, src_op);
+      Lw(scratch, src_op, trapper);
       vmv_vx(dst_v, scratch);
     } else if (memtype == MachineType::Int64()) {
       VU.set(kScratchReg, E64, m1);
-      LoadDouble(kScratchDoubleReg, src_op);
+      LoadDouble(kScratchDoubleReg, src_op, trapper);
       vfmv_vf(dst_v, kScratchDoubleReg);
     }
   }
+  if (protected_load_pc) {
+    DCHECK(InstructionAt(*protected_load_pc)->IsLoad());
+  }
 }
 
 void LiftoffAssembler::LoadLane(LiftoffRegister dst, LiftoffRegister src,
                                 Register addr, Register offset_reg,
                                 uintptr_t offset_imm, LoadType type,
                                 uint8_t laneidx, uint32_t* protected_load_pc,
-                                bool /* i64_offfset */) {
+                                bool /* i64_offset */) {
   UseScratchRegisterScope temps(this);
   Register scratch = temps.Acquire();
   MemOperand src_op = liftoff::GetMemOp(this, addr, offset_reg, offset_imm);
   MachineType mem_type = type.mem_type();
-  *protected_load_pc = pc_offset();
+  auto trapper = [protected_load_pc](int offset) {
+    if (protected_load_pc) *protected_load_pc = static_cast<uint32_t>(offset);
+  };
   if (mem_type == MachineType::Int8()) {
-    Lbu(scratch, src_op);
+    Lbu(scratch, src_op, trapper);
     VU.set(kScratchReg, E32, m1);
     li(kScratchReg, 0x1 << laneidx);
     vmv_sx(v0, kScratchReg);
     VU.set(kScratchReg, E8, m1);
     vmerge_vx(dst.fp().toV(), scratch, dst.fp().toV());
   } else if (mem_type == MachineType::Int16()) {
-    Lhu(scratch, src_op);
+    Lhu(scratch, src_op, trapper);
     VU.set(kScratchReg, E16, m1);
     li(kScratchReg, 0x1 << laneidx);
     vmv_sx(v0, kScratchReg);
     vmerge_vx(dst.fp().toV(), scratch, dst.fp().toV());
   } else if (mem_type == MachineType::Int32()) {
-    Lw(scratch, src_op);
+    Lw(scratch, src_op, trapper);
     VU.set(kScratchReg, E32, m1);
     li(kScratchReg, 0x1 << laneidx);
     vmv_sx(v0, kScratchReg);
     vmerge_vx(dst.fp().toV(), scratch, dst.fp().toV());
   } else if (mem_type == MachineType::Int64()) {
-    LoadDouble(kScratchDoubleReg, src_op);
+    LoadDouble(kScratchDoubleReg, src_op, trapper);
     VU.set(kScratchReg, E64, m1);
     li(kScratchReg, 0x1 << laneidx);
     vmv_sx(v0, kScratchReg);
@@ -1994,37 +1997,45 @@ void LiftoffAssembler::LoadLane(LiftoffRegister dst, LiftoffRegister src,
   } else {
     UNREACHABLE();
   }
+  if (protected_load_pc) {
+    DCHECK(InstructionAt(*protected_load_pc)->IsLoad());
+  }
 }
 
 void LiftoffAssembler::StoreLane(Register dst, Register offset,
                                  uintptr_t offset_imm, LiftoffRegister src,
                                  StoreType type, uint8_t lane,
                                  uint32_t* protected_store_pc,
-                                 bool /* i64_offfset */) {
+                                 bool /* i64_offset */) {
   MemOperand dst_op = liftoff::GetMemOp(this, dst, offset, offset_imm);
-  if (protected_store_pc) *protected_store_pc = pc_offset();
+  auto trapper = [protected_store_pc](int offset) {
+    if (protected_store_pc) *protected_store_pc = static_cast<uint32_t>(offset);
+  };
   MachineRepresentation rep = type.mem_rep();
   if (rep == MachineRepresentation::kWord8) {
     VU.set(kScratchReg, E8, m1);
     vslidedown_vi(kSimd128ScratchReg, src.fp().toV(), lane);
     vmv_xs(kScratchReg, kSimd128ScratchReg);
-    Sb(kScratchReg, dst_op);
+    Sb(kScratchReg, dst_op, trapper);
   } else if (rep == MachineRepresentation::kWord16) {
     VU.set(kScratchReg, E16, m1);
     vslidedown_vi(kSimd128ScratchReg, src.fp().toV(), lane);
     vmv_xs(kScratchReg, kSimd128ScratchReg);
-    Sh(kScratchReg, dst_op);
+    Sh(kScratchReg, dst_op, trapper);
   } else if (rep == MachineRepresentation::kWord32) {
     VU.set(kScratchReg, E32, m1);
     vslidedown_vi(kSimd128ScratchReg, src.fp().toV(), lane);
     vmv_xs(kScratchReg, kSimd128ScratchReg);
-    Sw(kScratchReg, dst_op);
+    Sw(kScratchReg, dst_op, trapper);
   } else {
     DCHECK_EQ(MachineRepresentation::kWord64, rep);
     VU.set(kScratchReg, E64, m1);
     vslidedown_vi(kSimd128ScratchReg, src.fp().toV(), lane);
     vfmv_fs(kScratchDoubleReg, kSimd128ScratchReg);
-    StoreDouble(kScratchDoubleReg, dst_op);
+    StoreDouble(kScratchDoubleReg, dst_op, trapper);
+  }
+  if (protected_store_pc) {
+    DCHECK(InstructionAt(*protected_store_pc)->IsStore());
   }
 }
 
@@ -2344,6 +2355,8 @@ void LiftoffStackSlots::Construct(int param_slots) {
   }
 }
 
+bool LiftoffAssembler::supports_f16_mem_access() { return false; }
+
 }  // namespace v8::internal::wasm
 
 #endif  // V8_WASM_BASELINE_RISCV_LIFTOFF_ASSEMBLER_RISCV32_INL_H_
