@@ -423,6 +423,15 @@ function _addHeaderLineDistinct(field, value, dest) {
   }
 }
 
+IncomingMessage.prototype._dumpAndCloseReadable = function _dumpAndCloseReadable() {
+  this._dumped = true;
+  this._readableState.ended = true;
+  this._readableState.endEmitted = true;
+  this._readableState.destroyed = true;
+  this._readableState.closed = true;
+  this._readableState.closeEmitted = true;
+};
+
 
 // Call this instead of resume() if we want to just
 // dump all the data to /dev/null
