@@ -25,6 +25,7 @@
 'use strict';
 
 const {
+  ArrayFromAsync,
   ArrayPrototypePush,
   BigIntPrototypeToString,
   Boolean,
@@ -33,6 +34,7 @@ const {
   ObjectDefineProperties,
   ObjectDefineProperty,
   Promise,
+  PromisePrototypeThen,
   PromiseResolve,
   ReflectApply,
   SafeMap,
@@ -40,6 +42,7 @@ const {
   StringPrototypeCharCodeAt,
   StringPrototypeIndexOf,
   StringPrototypeSlice,
+  SymbolDispose,
   uncurryThis,
 } = primordials;
 
@@ -85,6 +88,7 @@ const {
 const { toPathIfFileURL } = require('internal/url');
 const {
   customPromisifyArgs: kCustomPromisifyArgsSymbol,
+  deprecate,
   getLazy,
   kEmptyObject,
   promisify: {
@@ -142,13 +146,13 @@ const {
   validateFunction,
   validateInteger,
   validateObject,
+  validateOneOf,
   validateString,
   kValidateObjectAllowNullable,
 } = require('internal/validators');
 
 const permission = require('internal/process/permission');
 
-let truncateWarn = true;
 let fs;
 
 // Lazy loaded
@@ -158,7 +162,6 @@ let promises = null;
 let ReadStream;
 let WriteStream;
 let rimraf;
-let rimrafSync;
 let kResistStopPropagation;
 let ReadFileContext;
 
@@ -166,15 +169,10 @@ let ReadFileContext;
 // monkeypatching.
 let FileReadStream;
 let FileWriteStream;
+let Utf8Stream;
 
-function showTruncateDeprecation() {
-  if (truncateWarn) {
-    process.emitWarning(
-      'Using fs.truncate with a file descriptor is deprecated. Please use ' +
-      'fs.ftruncate with a file descriptor instead.',
-      'DeprecationWarning', 'DEP0081');
-    truncateWarn = false;
-  }
+function lazyLoadUtf8Stream() {
+  Utf8Stream ??= require('internal/streams/fast-utf8-stream');
 }
 
 // Ensure that callbacks run in the global context. Only use this function
@@ -269,12 +267,7 @@ ObjectDefineProperty(exists, kCustomPromisifiedSymbol, {
   },
 });
 
-// fs.existsSync never throws, it only returns true or false.
-// Since fs.existsSync never throws, users have established
-// the expectation that passing invalid arguments to it, even like
-// fs.existsSync(), would only get a false in return, so we cannot signal
-// validation errors to users properly out of compatibility concerns.
-// TODO(joyeecheung): deprecate the never-throw-on-invalid-arguments behavior
+let showExistsDeprecation = true;
 /**
  * Synchronously tests whether or not the given path exists.
  * @param {string | Buffer | URL} path
@@ -283,7 +276,13 @@ ObjectDefineProperty(exists, kCustomPromisifiedSymbol, {
 function existsSync(path) {
   try {
     path = getValidatedPath(path);
-  } catch {
+  } catch (err) {
+    if (showExistsDeprecation && err?.code === 'ERR_INVALID_ARG_TYPE') {
+      process.emitWarning(
+        'Passing invalid argument types to fs.existsSync is deprecated', 'DeprecationWarning', 'DEP0187',
+      );
+      showExistsDeprecation = false;
+    }
     return false;
   }
 
@@ -1030,10 +1029,6 @@ function renameSync(oldPath, newPath) {
  * @returns {void}
  */
 function truncate(path, len, callback) {
-  if (typeof path === 'number') {
-    showTruncateDeprecation();
-    return fs.ftruncate(path, len, callback);
-  }
   if (typeof len === 'function') {
     callback = len;
     len = 0;
@@ -1063,11 +1058,6 @@ function truncate(path, len, callback) {
  * @returns {void}
  */
 function truncateSync(path, len) {
-  if (typeof path === 'number') {
-    // legacy
-    showTruncateDeprecation();
-    return fs.ftruncateSync(path, len);
-  }
   if (len === undefined) {
     len = 0;
   }
@@ -1122,7 +1112,7 @@ function lazyLoadCp() {
 
 function lazyLoadRimraf() {
   if (rimraf === undefined)
-    ({ rimraf, rimrafSync } = require('internal/fs/rimraf'));
+    ({ rimraf } = require('internal/fs/rimraf'));
 }
 
 /**
@@ -1190,8 +1180,7 @@ function rmdirSync(path, options) {
     emitRecursiveRmdirWarning();
     options = validateRmOptionsSync(path, { ...options, force: false }, true);
     if (options !== false) {
-      lazyLoadRimraf();
-      return rimrafSync(path, options);
+      return binding.rmSync(path, options.maxRetries, options.recursive, options.retryDelay);
     }
   } else {
     validateRmdirOptions(options);
@@ -1242,11 +1231,8 @@ function rm(path, options, callback) {
  * @returns {void}
  */
 function rmSync(path, options) {
-  lazyLoadRimraf();
-  return rimrafSync(
-    getValidatedPath(path),
-    validateRmOptionsSync(path, options, false),
-  );
+  const opts = validateRmOptionsSync(path, options, false);
+  return binding.rmSync(getValidatedPath(path), opts.maxRetries, opts.recursive, opts.retryDelay);
 }
 
 /**
@@ -1802,13 +1788,17 @@ function readlinkSync(path, options) {
  * Creates the link called `path` pointing to `target`.
  * @param {string | Buffer | URL} target
  * @param {string | Buffer | URL} path
- * @param {string | null} [type_]
- * @param {(err?: Error) => any} callback_
+ * @param {string | null} [type]
+ * @param {(err?: Error) => any} callback
  * @returns {void}
  */
-function symlink(target, path, type_, callback_) {
-  const type = (typeof type_ === 'string' ? type_ : null);
-  const callback = makeCallback(arguments[arguments.length - 1]);
+function symlink(target, path, type, callback) {
+  if (callback === undefined) {
+    callback = makeCallback(type);
+    type = undefined;
+  } else {
+    validateOneOf(type, 'type', ['dir', 'file', 'junction', null, undefined]);
+  }
 
   // Due to the nature of Node.js runtime, symlinks has different edge cases that can bypass
   // the permission model security guarantees. Thus, this API is disabled unless fs.read
@@ -1821,7 +1811,7 @@ function symlink(target, path, type_, callback_) {
   target = getValidatedPath(target, 'target');
   path = getValidatedPath(path);
 
-  if (isWindows && type === null) {
+  if (isWindows && type == null) {
     let absoluteTarget;
     try {
       // Symlinks targets can be relative to the newly created path.
@@ -1871,8 +1861,8 @@ function symlink(target, path, type_, callback_) {
  * @returns {void}
  */
 function symlinkSync(target, path, type) {
-  type = (typeof type === 'string' ? type : null);
-  if (isWindows && type === null) {
+  validateOneOf(type, 'type', ['dir', 'file', 'junction', null, undefined]);
+  if (isWindows && type == null) {
     const absoluteTarget = pathModule.resolve(`${path}`, '..', `${target}`);
     if (statSync(absoluteTarget, { throwIfNoEntry: false })?.isDirectory()) {
       type = 'dir';
@@ -3061,6 +3051,36 @@ function mkdtempSync(prefix, options) {
   return binding.mkdtemp(prefix, options.encoding);
 }
 
+/**
+ * Synchronously creates a unique temporary directory.
+ * The returned value is a disposable object which removes the
+ * directory and its contents when disposed.
+ * @param {string | Buffer | URL} prefix
+ * @param {string | { encoding?: string; }} [options]
+ * @returns {object} A disposable object with a "path" property.
+ */
+function mkdtempDisposableSync(prefix, options) {
+  options = getOptions(options);
+
+  prefix = getValidatedPath(prefix, 'prefix');
+  warnOnNonPortableTemplate(prefix);
+
+  const path = binding.mkdtemp(prefix, options.encoding);
+  // Stash the full path in case of process.chdir()
+  const fullPath = pathModule.resolve(process.cwd(), path);
+
+  const remove = () => {
+    binding.rmSync(fullPath, 0 /* maxRetries */, true /* recursive */, 100 /* retryDelay */);
+  };
+  return {
+    path,
+    remove,
+    [SymbolDispose]() {
+      remove();
+    },
+  };
+}
+
 /**
  * Asynchronously copies `src` to `dest`. By
  * default, `dest` is overwritten if it already exists.
@@ -3204,18 +3224,11 @@ function glob(pattern, options, callback) {
   callback = makeCallback(callback);
 
   const Glob = lazyGlob();
-  // TODO: Use iterator helpers when available
-  (async () => {
-    try {
-      const res = [];
-      for await (const entry of new Glob(pattern, options).glob()) {
-        ArrayPrototypePush(res, entry);
-      }
-      callback(null, res);
-    } catch (err) {
-      callback(err);
-    }
-  })();
+  PromisePrototypeThen(
+    ArrayFromAsync(new Glob(pattern, options).glob()),
+    (res) => callback(null, res),
+    callback,
+  );
 }
 
 function globSync(pattern, options) {
@@ -3273,6 +3286,7 @@ module.exports = fs = {
   mkdirSync,
   mkdtemp,
   mkdtempSync,
+  mkdtempDisposableSync,
   open,
   openSync,
   openAsBlob,
@@ -3356,6 +3370,11 @@ module.exports = fs = {
     FileWriteStream = val;
   },
 
+  get Utf8Stream() {
+    lazyLoadUtf8Stream();
+    return Utf8Stream;
+  },
+
   // For tests
   _toUnixTimestamp: toUnixTimestamp,
 };
@@ -3367,10 +3386,50 @@ defineLazyProperties(
 );
 
 ObjectDefineProperties(fs, {
-  F_OK: { __proto__: null, enumerable: true, value: F_OK || 0 },
-  R_OK: { __proto__: null, enumerable: true, value: R_OK || 0 },
-  W_OK: { __proto__: null, enumerable: true, value: W_OK || 0 },
-  X_OK: { __proto__: null, enumerable: true, value: X_OK || 0 },
+  F_OK: {
+    __proto__: null,
+    enumerable: false,
+    get: deprecate(
+      function get() {
+        return F_OK || 0;
+      },
+      'fs.F_OK is deprecated, use fs.constants.F_OK instead',
+      'DEP0176',
+    ),
+  },
+  R_OK: {
+    __proto__: null,
+    enumerable: false,
+    get: deprecate(
+      function get() {
+        return R_OK || 0;
+      },
+      'fs.R_OK is deprecated, use fs.constants.R_OK instead',
+      'DEP0176',
+    ),
+  },
+  W_OK: {
+    __proto__: null,
+    enumerable: false,
+    get: deprecate(
+      function get() {
+        return W_OK || 0;
+      },
+      'fs.W_OK is deprecated, use fs.constants.W_OK instead',
+      'DEP0176',
+    ),
+  },
+  X_OK: {
+    __proto__: null,
+    enumerable: false,
+    get: deprecate(
+      function get() {
+        return X_OK || 0;
+      },
+      'fs.X_OK is deprecated, use fs.constants.X_OK instead',
+      'DEP0176',
+    ),
+  },
   constants: {
     __proto__: null,
     configurable: false,
