@@ -57,7 +57,7 @@ const {
 } = require('internal/errors').codes;
 const { maybeCacheSourceMap } = require('internal/source_map/source_map_cache');
 const moduleWrap = internalBinding('module_wrap');
-const { ModuleWrap } = moduleWrap;
+const { ModuleWrap, kEvaluationPhase } = moduleWrap;
 
 // Lazy-loading to avoid circular dependencies.
 let getSourceSync;
@@ -112,6 +112,7 @@ translators.set('module', function moduleStrategy(url, translateContext, parentU
   return module;
 });
 
+const { requestTypes: { kRequireInImportedCJS } } = require('internal/modules/esm/utils');
 /**
  * Loads a CommonJS module via the ESM Loader sync CommonJS translator.
  * This translator creates its own version of the `require` function passed into CommonJS modules.
@@ -131,7 +132,6 @@ function loadCJSModule(module, source, url, filename, isMain) {
   if (sourceMapURL) {
     maybeCacheSourceMap(url, source, module, false, sourceURL, sourceMapURL);
   }
-
   const cascadedLoader = require('internal/modules/esm/loader').getOrInitializeCascadedLoader();
   const __dirname = dirname(filename);
   // eslint-disable-next-line func-name-matching,func-style
@@ -154,7 +154,8 @@ function loadCJSModule(module, source, url, filename, isMain) {
 
     // FIXME(node:59666) Currently, the ESM loader re-invents require() here for imported CJS and this
     // requires a separate cache to be populated as well as introducing several quirks. This is not ideal.
-    const job = cascadedLoader.getModuleJobForRequireInImportedCJS(specifier, url, importAttributes);
+    const request = { specifier, attributes: importAttributes, phase: kEvaluationPhase, __proto__: null };
+    const job = cascadedLoader.getOrCreateModuleJob(url, request, kRequireInImportedCJS);
     job.runSync();
     let mod = cjsCache.get(job.url);
     assert(job.module, `Imported CJS module ${url} failed to load module ${job.url} using require() due to race condition`);
@@ -185,7 +186,9 @@ function loadCJSModule(module, source, url, filename, isMain) {
         specifier = `${pathToFileURL(path)}`;
       }
     }
-    const { url: resolvedURL } = cascadedLoader.resolveSync(specifier, url, kEmptyObject);
+
+    const request = { specifier, __proto__: null, attributes: kEmptyObject };
+    const { url: resolvedURL } = cascadedLoader.resolveSync(url, request);
     return urlToFilename(resolvedURL);
   });
   setOwnProperty(requireFn, 'main', process.mainModule);
