@@ -646,9 +646,8 @@ class MockTracker {
     // If the caller is already a file URL, use it as is. Otherwise, convert it.
     const hasFileProtocol = StringPrototypeStartsWith(filename, 'file://');
     const caller = hasFileProtocol ? filename : pathToFileURL(filename).href;
-    const { format, url } = sharedState.moduleLoader.resolveSync(
-      mockSpecifier, caller, kEmptyObject,
-    );
+    const request = { __proto__: null, specifier: mockSpecifier, attributes: kEmptyObject };
+    const { format, url } = sharedState.moduleLoader.resolveSync(caller, request);
     debug('module mock, url = "%s", format = "%s", caller = "%s"', url, format, caller);
     if (format) { // Format is not yet known for ambiguous files when detection is enabled.
       validateOneOf(format, 'format', kSupportedFormats);
