@@ -274,7 +274,7 @@ test('callback async throw after done', (t, done) => {
   done();
 });
 
-test('only is set but not in only mode', { only: true }, async (t) => {
+test('only is set on subtests but not in only mode', async (t) => {
   // All of these subtests should run.
   await t.test('running subtest 1');
   t.runOnly(true);
