@@ -2,7 +2,7 @@
 
 'use strict';
 const common = require('../../common');
-const addon = require(`./build/${common.buildType}/6_object_wrap`);
+const addon = require(`./build/${common.buildType}/myobject`);
 const { gcUntil } = require('../../common/gc');
 
 (function scope() {
