@@ -29,7 +29,7 @@ assert.throws(
 );
 
 (async () => {
-  const filehandle = await fsPromises.open(filepath, 'r');
+  await using filehandle = await fsPromises.open(filepath, 'r');
   assert.rejects(
     () => filehandle.read(buffer, 0, 1, 0),
     {
