@@ -27,7 +27,7 @@ server.listen(0, common.mustCall(() => {
   {
     const req = client.request();
 
-    req.on('response', () => {
+    req.on('response', common.mustCall(() => {
       // This one should be rejected because the server is over budget
       // on the current memory allocation
       const req = client.request();
@@ -40,7 +40,7 @@ server.listen(0, common.mustCall(() => {
         server.close();
         client.destroy();
       }));
-    });
+    }));
 
     req.resume();
     req.on('close', common.mustCall());
